#!/usr/bin/env python
# -*- coding: utf-8 -*-
'''
utils tests
'''
import pytest

from pydna import eq, read, Dseqrecord
from pydna.utils import shift_origin
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord



def test_eq():
    ''' test eq'''
    assert   eq( "AAA" ,"TTT", linear   = True )
    assert  eq( "AAA" ,"TTT", linear   = False)

    assert  eq( "aAA" ,"TtT", linear   = True )
    assert  eq( "AAa" ,"TtT", linear   = False)


    assert  eq( "ATA" ,"AAT", circular = True )
    assert not eq( "ATA" ,"AAT", circular = False)
    assert  eq( "AAA" ,"AAA", linear   = True ) 
    assert  eq( "AAA" ,"AAA", linear   = False) 

    assert  eq( "ATA" ,Seq("AAT"), circular = True ) 
    assert not eq( "ATA" ,Seq("AAT"), circular = False) 
    assert  eq( "AAA" ,Seq("AAA"), linear   = True ) 
    assert  eq( "AAA" ,Seq("AAA"), linear   = False) 

    assert  eq( "ATA" ,SeqRecord("AAT"), circular = True ) 
    assert not eq( "ATA" ,SeqRecord("AAT"), circular = False)
    assert  eq( "AAA" ,SeqRecord("AAA"), linear   = True ) 
    assert  eq( "AAA" ,SeqRecord("AAA"), linear   = False) 

    assert  eq( "ATA" ,Dseqrecord("AAT"), circular = True ) 
    assert not eq( "ATA" ,Dseqrecord("AAT"), circular = False) 
    assert  eq( "AAA" ,Dseqrecord("AAA"), linear   = True ) 
    assert  eq( "AAA" ,Dseqrecord("AAA"), linear   = False) 

    assert  eq( Seq("ATA") ,SeqRecord("AAT"), circular = True )
    assert not eq( Seq("ATA") ,SeqRecord("AAT"), circular = False)
    assert  eq( Seq("AAA") ,SeqRecord("AAA"), linear   = True )
    assert  eq( Seq("AAA") ,SeqRecord("AAA"), linear   = False)

    assert  eq( Seq("ATA") ,Dseqrecord("AAT"), circular = True )
    assert not eq( Seq("ATA") ,Dseqrecord("AAT"), circular = False)
    assert  eq( Seq("AAA") ,Dseqrecord("AAA"), linear   = True ) 
    assert  eq( Seq("AAA") ,Dseqrecord("AAA"), linear   = False) 

    assert  eq( Dseqrecord("AAA",circular=False) ,Dseqrecord("AAA",circular=False)) 
    assert  eq( Dseqrecord("AAA",circular=True)  ,Dseqrecord("AAA",circular=True))   
    assert not eq( Dseqrecord("ATA",circular=False) ,Dseqrecord("AAT",circular=False)) 
    assert  eq( Dseqrecord("ATA",circular=True)  ,Dseqrecord("AAT",circular=True)) 


def test_shift_origin():

    pCAPs   = read("pCAPs.gb")
    assert pCAPs.circular
    pCAPs_b = shift_origin(pCAPs, 200)
    assert len(pCAPs) == len(pCAPs_b)
    assert pCAPs_b.circular
    assert eq(pCAPs, pCAPs_b)
    pCAPs_b_linear = pCAPs_b.tolinear()
    assert eq(pCAPs, pCAPs_b_linear, circular=True)
    pCAPs_c = pCAPs[200:]+pCAPs[:200]
    assert eq(pCAPs, pCAPs_c, circular=True)
    #with self.assertRaisesRegex(ValueError, "shift"):
    #    pCAPs_b = shift_origin(pCAPs, 20000)


def test_copy_features():

    from pydna.utils import seguid, copy_features
    from pydna import read
    a=read("pCAPs.gb")
    b=read("pCAPs_fasta.txt")

    for sh in [1,2,3,3127,3128,3129]:
        newb = (b[sh:]+b[:sh]).looped()
        copy_features(a, newb)
        #print "a",[len(str(f.extract(a).seq.lower()) for f in a.features if len(f)>10]
        #print "b",[len(str(f.extract(newb).seq).lower()) for f in newb.features]
        x= sorted([str(f.extract(a).seq).lower() for f in a.features if len(f)>10],key=len)
        y= sorted([str(f.extract(newb).seq).lower() for f in newb.features],key=len)
        assert x==y

    b=b.rc()

    for sh in [1,2,3,3127,3128,3129]:
        newb = b[sh:]+b[:sh]
        copy_features(a, newb)
        
        
        x = sorted([str(f.extract(a).seq).lower() for f in a.features if len(f)>10],key=len)
        y = sorted([str(f.extract(newb).seq).lower() for f in newb.features],key=len)
        assert x==y

    seguid_bla = "riT98j2v4NxVS8sbw_Q8epCwQwo"
    seguid_cre = "xLZ2xs2O8CUMmWh2OrhmNFp5ZLg"

    copy_features(a, b)
    assert [seguid(f.extract(b).seq) for f in b.features] == [seguid_cre, seguid_cre, seguid_bla, seguid_bla]

    b=read("pCAPs_fasta.txt").looped()

    b=b.synced("attaacgagtgccgtaaacgacgatggttttacc")

    copy_features(a, b)
    assert [seguid(f.extract(b).seq) for f in b.features] == [seguid_cre,seguid_cre,seguid_bla,seguid_bla]

    b=read("pCAPs_fasta.txt").looped()
    b=b.synced("ttaacgagtgccgtaaacgacgatggttttacc")

    copy_features(a, b)
    assert [seguid(f.extract(b).seq) for f in b.features] == [seguid_cre,seguid_cre,seguid_bla,seguid_bla]

    b=read("pCAPs_fasta.txt").looped()
    b=b.synced("taacgagtgccgtaaacgacgatggttttacc")

    copy_features(a, b)
    assert [seguid(f.extract(b).seq) for f in b.features] == [seguid_bla,seguid_bla]

    b=read("pCAPs_fasta.txt").looped()
    b=b.synced("gttaccaatgcttaatcagtgaggcacctatctcagc")

    copy_features(a, b)
    assert [seguid(f.extract(b).seq) for f in b.features] == [seguid_cre,seguid_cre,seguid_bla,seguid_bla]

    b=read("pCAPs_fasta.txt").looped()
    b=b.synced("ttaccaatgcttaatcagtgaggcacctatctcagc")

    copy_features(a, b)
    assert [seguid(f.extract(b).seq) for f in b.features] == [seguid_cre,seguid_cre,seguid_bla,seguid_bla]

    b=read("pCAPs_fasta.txt").looped()
    b=b.synced("taccaatgcttaatcagtgaggcacctatctcagc")

    copy_features(a, b)
    assert [seguid(f.extract(b).seq) for f in b.features] == [seguid_cre,seguid_cre,]


def test_cseguid():
    from pydna.utils import cseguid
    x="tcgcgcgtttcggtgatgacggtgaaaacctctgacacatgcagctcccggagacggtcacagcttgtctgtaagcggatgccgggagcagacaagcccgtcagggcgcgtcagcgggtgttggcgggtgtcggggctggcttaactatgcggcatcagagcagattgtactgagagtgcaccatatgcggtgtgaaataccgcacagatgcgtaaggagaaaataccgcatcaggcgccattcgccattcaggctgcgcaactgttgggaagggcgatcggtgcgggcctcttcgctattacgccagctggcgaaagggggatgtgctgcaaggcgattaagttgggtaacgccagggttttcccagtcacgacgttgtaaaacgacggccagtgaattcgagctcggtacccgggGATCTATGAATATGGATCCGACTTACTGCAGGAATTCAAGCTACTGTTAGAgatcctctagagtcgacctgcaggcatgcaagcttggcgtaatcatggtcatagctgtttcctgtgtgaaattgttatccgctcacaattccacacaacatacgagccggaagcataaagtgtaaagcctggggtgcctaatgagtgagctaactcacattaattgcgttgcgctcactgcccgctttccagtcgggaaacctgtcgtgccagctgcattaatgaatcggccaacgcgcggggagaggcggtttgcgtattgggcgctcttccgcttcctcgctcactgactcgctgcgctcggtcgttcggctgcggcgagcggtatcagctcactcaaaggcggtaatacggttatccacagaatcaggggataacgcaggaaagaacatgtgagcaaaaggccagcaaaaggccaggaaccgtaaaaaggccgcgttgctggcgtttttccataggctccgcccccctgacgagcatcacaaaaatcgacgctcaagtcagaggtggcgaaacccgacaggactataaagataccaggcgtttccccctggaagctccctcgtgcgctctcctgttccgaccctgccgcttaccggatacctgtccgcctttctcccttcgggaagcgtggcgctttctcatagctcacgctgtaggtatctcagttcggtgtaggtcgttcgctccaagctgggctgtgtgcacgaaccccccgttcagcccgaccgctgcgccttatccggtaactatcgtcttgagtccaacccggtaagacacgacttatcgccactggcagcagccactggtaacaggattagcagagcgaggtatgtaggcggtgctacagagttcttgaagtggtggcctaactacggctacactagaagaacagtatttggtatctgcgctctgctgaagccagttaccttcggaaaaagagttggtagctcttgatccggcaaacaaaccaccgctggtagcggtggtttttttgtttgcaagcagcagattacgcgcagaaaaaaaggatctcaagaagatcctttgatcttttctacggggtctgacgctcagtggaacgaaaactcacgttaagggattttggtcatgagattatcaaaaaggatcttcacctagatccttttaaattaaaaatgaagttttaaatcaatctaaagtatatatgagtaaacttggtctgacagttaccaatgcttaatcagtgaggcacctatctcagcgatctgtctatttcgttcatccatagttgcctgactccccgtcgtgtagataactacgatacgggagggcttaccatctggccccagtgctgcaatgataccgcgagacccacgctcaccggctccagatttatcagcaataaaccagccagccggaagggccgagcgcagaagtggtcctgcaactttatccgcctccatccagtctattaattgttgccgggaagctagagtaagtagttcgccagttaatagtttgcgcaacgttgttgccattgctacaggcatcgtggtgtcacgctcgtcgtttggtatggcttcattcagctccggttcccaacgatcaaggcgagttacatgatcccccatgttgtgcaaaaaagcggttagctccttcggtcctccgatcgttgtcagaagtaagttggccgcagtgttatcactcatggttatggcagcactgcataattctcttactgtcatgccatccgtaagatgcttttctgtgactggtgagtactcaaccaagtcattctgagaatagtgtatgcggcgaccgagttgctcttgcccggcgtcaatacgggataataccgcgccacatagcagaactttaaaagtgctcatcattggaaaacgttcttcggggcgaaaactctcaaggatcttaccgctgttgagatccagttcgatgtaacccactcgtgcacccaactgatcttcagcatcttttactttcaccagcgtttctgggtgagcaaaaacaggaaggcaaaatgccgcaaaaaagggaataagggcgacacggaaatgttgaatactcatactcttcctttttcaatattattgaagcatttatcagggttattgtctcatgagcggatacatatttgaatgtatttagaaaaataaacaaataggggttccgcgcacatttccccgaaaagtgccacctgacgtctaagaaaccattattatcatgacattaacctataaaaataggcgtatcacgaggccctttcgtc"
    assert cseguid(x) == cseguid(x.upper()) == cseguid(x.lower()) == 'JgiKgBksd2v3q99NStKLepoQCm8'
    from Bio.SeqUtils.CheckSum  import seguid as base64_seguid



if __name__ == '__main__':
    pytest.cmdline.main([__file__, "-v", "-s"])



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import pytest
import pydna

def test_efetch_download_text():
    # see https://www.ncbi.nlm.nih.gov/books/NBK25499/#chapter4.EFetch
    url = "https://eutils.ncbi.nlm.nih.gov/entrez/eutils/efetch.fcgi?db=nuccore&id=E05006&strand=1&rettype=gb&retmode=gbwithparts"
    cachevar = os.environ["pydna_cache"]
    gbdata = pydna.download_text(url)
    os.environ["pydna_cache"] = cachevar
    with open("E05006.gb") as f:
        localdata = f.read().strip()
    assert localdata==gbdata

def test_biopython_download():
    from Bio import Entrez
    from Bio import SeqIO    
    Entrez.email = "bjornjobb@gmail.com"    
    handle = Entrez.efetch(db="nuccore",
                           id="E05006",
                           rettype="gb",
                           retmode="text")    
    result = SeqIO.read(handle, "genbank")    
    assert str(result.seq) == "ATATGGGTACCGATCGTACGGACCA"

def test_pydna_download_fresh():
    cachevar = os.environ["pydna_cache"]
    os.environ["pydna_cache"] = "nocache"
    gb = pydna.Genbank("bjornjobb@gmail.com")
    result = gb.nucleotide("E05006")
    assert len(result) == 25
    assert str(result.seq) == "ATATGGGTACCGATCGTACGGACCA"
    os.environ["pydna_cache"] = cachevar
    
def test_pydna_download_cache():
    cachevar = os.environ["pydna_cache"]
    os.environ["pydna_cache"] = "cached"
    gb = pydna.Genbank("bjornjobb@gmail.com")
    result = gb.nucleotide("E05006")
    assert len(result) == 25
    assert str(result.seq) == "ATATGGGTACCGATCGTACGGACCA"
    result = gb.nucleotide("E05006")
    assert len(result) == 25
    assert str(result.seq) == "ATATGGGTACCGATCGTACGGACCA"    
    os.environ["pydna_cache"] = cachevar 
    

if __name__ == '__main__':
    pytest.cmdline.main([__file__, "-v", "-s"])
    

    



#!/usr/bin/env python
# -*- coding: utf-8 -*-
'''
test YEp24PGK_XK
'''
import pytest
import sys
import pydna

def test_empty():
    ''' test YEp24PGK_XK'''       
    p1 =   pydna.read("primer1.txt", ds = False)
    p3 =   pydna.read("primer3.txt", ds = False)
    XKS1 = pydna.read("XKS1_orf.txt")
    YEp24PGK = pydna.read("YEp24PGK.txt")

    PCR_prod = pydna.pcr(p1, p3, XKS1)

    from Bio.Restriction import BamHI
    
    stuffer1, insert, stuffer2 = PCR_prod.cut(BamHI)

    from Bio.Restriction import BglII

    YEp24PGK_BglII = YEp24PGK.cut(BglII).pop()

    YEp24PGK_XK = YEp24PGK_BglII + insert

    YEp24PGK_XK=YEp24PGK_XK.looped()

    YEp24PGK_XK = YEp24PGK_XK.synced("gaattctgaaccagtcctaaaacgagtaaataggaccggcaattc") #YEp24PGK)

    assert YEp24PGK_XK.seguid()  == "HRVpCEKWcFsKhw_W-25ednUfldI"
    assert YEp24PGK_XK.cseguid() == "t9fs_9UvEuD-Ankyy8XEr1hD5DQ"

    YEp24PGK_XK_correct = pydna.read("YEp24PGK_XK_manually_assembled.txt")
    assert YEp24PGK_XK_correct.cseguid() == "t9fs_9UvEuD-Ankyy8XEr1hD5DQ"
    assert pydna.eq(YEp24PGK_XK, YEp24PGK_XK_correct)

if __name__ == '__main__':
    pytest.cmdline.main([__file__, "-v", "-s"])













#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright 2013 by Björn Johansson.  All rights reserved.
# This code is part of the Python-dna distribution and governed by its
# license.  Please see the LICENSE.txt file that should have been included
# as part of this package.

'''Provides two classes, Dseq and Dseqrecord, for handling double stranded
DNA sequences. Dseq and Dseqrecord are subclasses of Biopythons
Seq and SeqRecord classes, respectively. These classes support the
notion of circular and linear DNA.

'''
import os
import re
import io
import textwrap
import glob

from Bio                    import SeqIO
from Bio.Alphabet.IUPAC     import IUPACAmbiguousDNA
from Bio.GenBank            import RecordParser
from .genbankfile            import GenbankFile
from .dseqrecord             import Dseqrecord
from .primer                 import Primer  

def parse2(data, ds = True):
    '''experimental'''

    pattern =  r"(?:>.+\n^(?:^[^>]+?)(?=\n\n|>|LOCUS|ID))|(?:(?:LOCUS|ID)(?:(?:.|\n)+?)^//)"

    def extract_seqs(raw):
        raw = textwrap.dedent(raw).strip()
        raw = raw.replace( '\r\n', '\n')
        raw = raw.replace( '\r',   '\n')
        return re.findall(pattern, textwrap.dedent(raw+ "\n\n"),re.MULTILINE)

    files=[]
    rawseqs=[]

    if not hasattr(data, '__iter__'):
        data = (data,)
    for item in data:
        for pth in glob.glob(item):
            if os.path.isfile(pth):
                files.append(os.path.abspath(pth))
            else:
                for dirpath,_,filenames in os.walk(pth):
                    for f in filenames:
                        files.append( os.path.abspath(os.path.join(dirpath, f)))
            for file_ in files:
                with open(file_,'r') as f:
                    rawseqs.extend(extract_seqs(f.read()))
            files=[]
        else:
            rawseqs.extend(extract_seqs(item))
    sequences = []
    while rawseqs:
        circular = False
        rawseq = rawseqs.pop(0)
        handle = io.StringIO(rawseq)
        try:
            parsed = SeqIO.read(handle, "embl", alphabet=IUPACAmbiguousDNA())
            #original_format = "embl"
            if "circular" in rawseq.splitlines()[0]:
                circular = True
        except ValueError:
            handle.seek(0)
            try:
                parsed = SeqIO.read(handle, "genbank", alphabet=IUPACAmbiguousDNA())
                #original_format = "genbank"
                handle.seek(0)
                parser = RecordParser()
                residue_type = parser.parse(handle).residue_type
                if "circular" in residue_type:
                    circular = True
            except ValueError:
                handle.seek(0)
                try:
                    parsed = SeqIO.read(handle, "fasta", alphabet=IUPACAmbiguousDNA())
                    if "circular" in rawseq.splitlines()[0]:
                        circular = True
                except ValueError:
                    continue

        if ds:
            sequences.append(Dseqrecord(parsed, circular = circular))
        else:
            sequences.append(parsed)
        handle.close()
    return sequences

def parse(data, ds = True):
    '''This function returns *all* DNA sequences found in data. If no
    sequences are found, an empty list is returned. This is a greedy
    function, use carefully.

    Parameters
    ----------
    data : string or iterable
        The data parameter is a string containing:

        1. an absolute path to a local file.
           The file will be read in text
           mode and parsed for EMBL, FASTA
           and Genbank sequences.

        2. a string containing one or more
           sequences in EMBL, GENBANK,
           or FASTA format. Mixed formats
           are allowed.

        3. data can be a list or other iterable where the elements are 1 or 2

    ds : bool
        If True double stranded :class:`Dseqrecord` objects are returned.
        If False single stranded :class:`Bio.SeqRecord` [#]_ objects are returned.

    Returns
    -------
    list
        contains Dseqrecord or SeqRecord objects

    References
    ----------

    .. [#] http://biopython.org/wiki/SeqRecord

    See Also
    --------
    read

    '''
    
    def embl_gb_fasta(raw, ds, path=None):
        
        pattern =  r"(?:>.+\n^(?:^[^>]+?)(?=\n\n|>|LOCUS|ID))|(?:(?:LOCUS|ID)(?:(?:.|\n)+?)^//)"
        
        result_list = []
        
        rawseqs = re.findall(pattern, textwrap.dedent(raw + "\n\n"), flags=re.MULTILINE)
        
        for rawseq in rawseqs:
            handle = io.StringIO(rawseq)
            circular=False
            try:
                parsed = SeqIO.read(handle, "embl", alphabet=IUPACAmbiguousDNA())
                if "circular" in rawseq.splitlines()[0]:
                    circular = True
            except ValueError:
                handle.seek(0)
                try:
                    parsed = SeqIO.read(handle, "genbank", alphabet=IUPACAmbiguousDNA())
                    handle.seek(0)
                    parser = RecordParser()
                    residue_type = parser.parse(handle).residue_type
                    if "circular" in residue_type:
                        circular = True
                except ValueError:
                    handle.seek(0)
                    try:
                        parsed = SeqIO.read(handle, "fasta", alphabet=IUPACAmbiguousDNA())
                        if "circular" in rawseq.splitlines()[0]:
                            circular = True
                    except ValueError:
                        parsed = ""
            handle.close()
            
            if ds and path:
                result_list.append( GenbankFile(parsed, circular=circular, path=path) )
            elif ds:
                result_list.append( Dseqrecord(parsed, circular =circular) )
            else:
                result_list.append( parsed )
                
        return result_list

    # a string is an iterable datatype but on Python2.x it doesn't have an __iter__ method.
    if not hasattr(data, '__iter__') or isinstance(data, (str, bytes)):
        data = (data,)
        
    sequences = []
    
    for item in data:
        try:
            # item is a path to a utf-8 encoded text file?
            with open(item, 'r', encoding="utf-8") as f:    
                raw = f.read()  
        except IOError:
            # item was not a path, add sequences parsed from item
            raw=item
            path=None
        else:   
            # item was a readable text file, seqences are parsed from the file
            path = item
        finally:
            sequences.extend( embl_gb_fasta(raw, ds, path) )
    return sequences
    
def parse_primers(data):
    return [Primer(x) for x in parse(data, ds=False)]

    
if __name__=="__main__":
    import os
    cache = os.getenv("pydna_cache")
    os.environ["pydna_cache"]="nocache"
    import doctest
    doctest.testmod()
    os.environ["pydna_cache"]=cache



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import math

def tmstaluc98(primer, dnac=50, saltc=50, **kwargs):
    '''Returns the melting temperature (Tm) of the primer using
    the nearest neighbour algorithm. Formula and thermodynamic data
    is taken from SantaLucia 1998 [1]_. This implementation gives the same
    answer as the one provided by Biopython (See Examples).

    Thermodynamic data used:

    =====  ====  ====
    pair   dH    dS
    =====  ====  ====
    AA/TT  7.9   22.2
    AT/TA  7.2   20.4
    TA/AT  7.2   21.3
    CA/GT  8.5   22.7
    GT/CA  8.4   22.4
    CT/GA  7.8   21.0
    GA/CT  8.2   22.2
    CG/GC  10.6  27.2
    GC/CG  9.8   24.4
    GG/CC  8.0   19.9
    =====  ====  ====

    Parameters
    ----------
    primer : string
        Primer sequence 5'-3' in UPPERCASE

    Returns
    -------
    tm : float
        tm of the primer


    Examples
    --------

    >>> from pydna.tm import tmstaluc98
    >>> from Bio.SeqUtils.MeltingTemp import Tm_staluc
    >>> tmstaluc98("ACGTCATCGACACTATCATCGAC")
    54.55597724052518
    >>>



    '''

    nntermsl={  "AA": (7.9  , 22.2),
                "TT": (7.9  , 22.2),
                "AT": (7.2  , 20.4),
                "TA": (7.2  , 21.3),
                "CA": (8.5  , 22.7),
                "TG": (8.5  , 22.7),
                "GT": (8.4  , 22.4),
                "AC": (8.4  , 22.4),
                "CT": (7.8  , 21.0),
                "AG": (7.8  , 21.0),
                "GA": (8.2  , 22.2),
                "TC": (8.2  , 22.2),
                "CG": (10.6 , 27.2),
                "GC": (9.8  , 24.4),
                "GG": (8    , 19.9),
                "CC": (8    , 19.9),
                "A" : (0    , 0   ),
                "C" : (0    , 0   ),
                "G" : (0    , 0   ),
                "T" : (0    , 0   )  }

    helixinit = {   "G": (-0.1 ,2.8),
                    "C": (-0.1 ,2.8),
                    "A": (-2.3, -4.1),
                    "T": (-2.3, -4.1) }

    dH, dS = helixinit[primer[0]]
    H ,  S = helixinit[primer[-1]]
    dH = dH+H
    dS = dS+S

    for p in range(len(primer)):
        dn = primer[p:p+2]
        H,S = nntermsl[dn]
        dH+=H
        dS+=S
    R = 1.987 # universal gas constant in Cal/degrees C*Mol
    k = (dnac/4.0)*1e-9
    dS = dS-0.368*(len(primer)-1)*math.log(float(saltc)/1e3)
    tm = ((1000* (-dH))/(-dS+(R * (math.log(k)))))-273.15
    return tm

def tmbreslauer86(primer, dnac=500.0, saltc=50,thermodynamics=False):
    '''Returns the melting temperature (Tm) of the primer using
    the nearest neighbour algorithm. Formula and thermodynamic data
    is taken from Breslauer 1986.

    These data are no longer widely used.


    Breslauer 1986, table 2 [#]_

    =====  ===== ====   ===
    pair   dH    dS     dG
    =====  ===== ====   ===
    AA/TT  9.1   24.0   1.9
    AT/TA  8.6   23.9   1.5
    TA/AT  6.0   16.9   0.9
    CA/GT  5.8   12.9   1.9
    GT/CA  6.5   17.3   1.3
    CT/GA  7.8   20.8   1.6
    GA/CT  5.6   13.5   1.6
    CG/GC  11.9  27.8   3.6
    GC/CG  11.1  26.7   3.1
    GG/CC  11.0  26.6   3.1
    =====  ===== ====   ===

    Parameters
    ----------
    primer : string
        Primer sequence 5'-3' in UPPERCASE

    Returns
    -------
    tm : float


    References
    ----------
    .. [#] K.J. Breslauer et al., “Predicting DNA Duplex Stability from the Base Sequence,” Proceedings of the National Academy of Sciences 83, no. 11 (1986): 3746.


    Examples
    ---------

    >>> from pydna.tm import tmbreslauer86
    >>> tmbreslauer86("ACGTCATCGACACTATCATCGAC")
    64.28863985851899


    '''

    nntermbr={  "AA": (9.1   ,24.0   ,1.9),
                "TT": (9.1   ,24.0   ,1.9),
                "AT": (8.6   ,23.9   ,1.5),
                "TA": (6.0   ,16.9   ,0.9),
                "CA": (5.8   ,12.9   ,1.9),
                "TG": (5.8   ,12.9   ,1.9),
                "GT": (6.5   ,17.3   ,1.3),
                "AC": (6.5   ,17.3   ,1.3),
                "CT": (7.8   ,20.8   ,1.6),
                "AG": (7.8   ,20.8   ,1.6),
                "GA": (5.6   ,13.5   ,1.6),
                "TC": (5.6   ,13.5   ,1.6),
                "CG": (11.9  ,27.8   ,3.6),
                "GC": (11.1  ,26.7   ,3.1),
                "GG": (11.0  ,26.6   ,3.1),
                "CC": (11.0  ,26.6   ,3.1),
                "A" : (0     , 0     ,0),
                "C" : (0     , 0     ,0),
                "G" : (0     , 0     ,0),
                "T" : (0     , 0     ,0),     }
    dH=3.4
    dS=12.4
    dG=0
    for p in range(len(primer)):
        dn = primer[p:p+2]
        H,S,G = nntermbr[dn]
        dG+=G
        dH+=H
        dS+=S

    R = 1.9872          # universal gas constant in Cal/degrees C*Mol
    k = dnac*1E-9/2.0
    dH = dH - 5
    dS = dS-0.368*(len(primer)-1)*math.log(float(saltc)/1E3) # SantaLucia salt correction formula
    tm = 1000 * -dH /(-dS + R * math.log(k) )  - 273.15 # degrees Celsius

    if thermodynamics:
        return tm,dH,dS
    else:
        return tm


def tmbresluc(primer, primerc=500.0, saltc=50, thermodynamics=False):
    '''Returns the tm for a primer using a formula adapted to polymerases
    with a DNA binding domain, such as the Phusion polymerase.

    Parameters
    ----------

    primer : string
        primer sequence 5'-3'

    primerc : float
       primer concentration in nM), set to 500.0 nm by default.

    saltc : float, optional
       Monovalent cation concentration in mM, set to 50.0 mM by default.

    thermodynamics : bool, optional
        prints details of the thermodynamic data to stdout. For
        debugging only.

    Returns
    -------
    tm : float
        the tm of the primer

    tm,dH,dS : tuple
        tm and dH and dS used for the calculation

    '''

    from . import _thermodynamic_data

    saltc = float(saltc)/1000
    pri  = primerc/10E7
    dS = -12.4
    dH = -3400

    STR = primer.lower();

    for i in range(len(STR)-1):
        n1=ord(STR[i])
        n2=ord(STR[i+1])
        dH += _thermodynamic_data.dHBr[n1 - 97][n2 - 97]
        dS += _thermodynamic_data.dSBr[n1 - 97][n2 - 97]

    tm = (dH / (1.9872 * math.log(pri / 1600) + dS) + (16.6 * math.log(saltc)) / math.log(10)) - 273.15

    if thermodynamics:
        return tm,dH,dS
    else:
        return tm

def basictm(primer, *args, **kwargs):
    '''Returns the melting temperature (Tm) of the primer using
    the basic formula. This function returns the same value as
    the Biopython Bio.SeqUtils.MeltingTemp.Tm_Wallace

    | Tm = (wA+xT)*2 + (yG+zC)*4 assumed 50mM monovalent cations
    |
    | w = number of A in primer
    | x = number of T in primer
    | y = number of G in primer
    | z = number of C in primer

    Parameters
    ----------
    primer : string
        Primer sequence 5'-3'

    Returns
    -------
    tm : int

    Examples
    --------
    >>> from pydna.tm import basictm
    >>> basictm("ggatcc")
    20
    >>>

    '''
    primer = str(primer).lower()
    return (primer.count("a") + primer.count("t"))*2 + (primer.count("g") + primer.count("c"))*4

# http://www.promega.com/techserv/tools/biomath/calc11.htm#melt_results        
        

def Q5():
    '''
    for Q5 Ta they take the lower of the two Tms and add 1C 
    (up to 72C). For Phusion they take the lower of the two 
    and add 3C (up to 72C). 
    '''
    pass


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright 2013, 2014 by Björn Johansson.  All rights reserved.
# This code is part of the Python-dna distribution and governed by its
# license.  Please see the LICENSE.txt file that should have been included
# as part of this package.

'''
This module contain functions for primer design.

'''
from Bio.Alphabet.IUPAC import IUPACAmbiguousDNA
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
from .tm import tmbresluc

class Primer(SeqRecord):
    '''This class holds information about a primer and its position on a template '''
    def __init__(self, record, *args,
                 position=None, footprint=None, tail=None, concentration = 1000.0, **kwargs):
        if hasattr(record, "features"):            
            for key, value in list(record.__dict__.items()):
                setattr(self, key, value )
        elif hasattr(record, "alphabet"):
            super().__init__(record, *args, **kwargs)            
        else:        
            super().__init__(Seq(record, IUPACAmbiguousDNA), *args, **kwargs)
        self.position  = position
        self.footprint = footprint
        self.tail      = tail
        self.concentration = concentration
    def __repr__(self):        
        return "{id} {len}-mer:5'{seq}-3'".format(id=self.id,len=len(self),seq=self.seq)
    def __radd__(self, other):
        new = super().__radd__(other)
        return Primer(new.seq)
    def tm(self, saltc=50.0, formula=tmbresluc):
        return formula( str(self.seq).upper(), primerc=self.concentration, saltc=saltc )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
'''
test parse
'''

import pytest
import sys
import os

from pydna import pcr, Assembly, Dseqrecord, assembly_primers, cloning_primers, parse

from textwrap import dedent

def test_primer_Design():
    ''' test_primer_design'''

    a=Dseqrecord("atgactgctaacccttccttggtgttgaacaagatcgacgacatttcgttcgaaacttacgatg")
    b=Dseqrecord("ccaaacccaccaggtaccttatgtaagtacttcaagtcgccagaagacttcttggtcaagttgcc")
    c=Dseqrecord("tgtactggtgctgaaccttgtatcaagttgggtgttgacgccattgccccaggtggtcgtttcgtt")

    primer_pairs = assembly_primers([a,b,c])

    frags=[]

    for (f,r),t in zip(primer_pairs,[a,b,c]):
        frags.append(pcr(f,r,t))

    asm=Assembly(frags)

    assert asm.linear_products[0].seguid() == "1eNv3d_1PqDPP8qJZIVoA45Www8"

    frags=[]

    primer_pairs = assembly_primers([a,b,c], circular=True)

    for (f,r),t in zip(primer_pairs,[a,b,c]):
        frags.append(pcr(f,r,t))

    #print frags

    asm=Assembly(frags)

    assert asm.circular_products[0].cseguid() == "V3Mi8zilejgyoH833UbjJOtDMbc"



def test_primer_Design_with_linker():
    ''' test_primer_design'''


    from pydna import Dseqrecord, Assembly, pcr, assembly_primers

    b  = Dseqrecord("agctactgactattaggggttattctgatcatctgatctactatctgactgtactgatcta")
    l  = Dseqrecord("AAATTTCCCGGG")
    c  = Dseqrecord("tctgatctactatctgactgtactgatctattgacactgtgatcattctagtgtattactc")

    ((bf,br),(cf,cr)) = assembly_primers((b,l,c))

    nb = pcr((bf,br),b)
    nc = pcr((cf,cr),c)

    asm1 = Assembly((nb,nc))

    assert asm1.linear_products[0].seguid(),(b+l+c).seguid() == 'l95igKB8iKAKrvvqE9CYksyNx40'


    b  = Dseqrecord("agctactgactattaggggttattctgatcatctgatctactatctgactgtactgatcta")
    l  = Dseqrecord("AAATTTCCCGGG")
    c  = Dseqrecord("tctgatctactatctgactgtactgatctattgacactgtgatcattctagtgtattactc")

    ((bf,br),(cf,cr)) = assembly_primers((b,l,c), circular = True)

    nb = pcr((bf,br),b)
    nc = pcr((cf,cr),c)

    asm = Assembly((nb,nc))

    #print (b+l+c).looped().seq

    assert (b+l+c).looped().cseguid() == asm.circular_products[0].cseguid()
    #print (b+l+c).looped().cseguid() == 'jdHXfQI5k4Sk2ESiZYfKv4oP2FI'

    assert (b+l+c).looped().cseguid() == 'jdHXfQI5k4Sk2ESiZYfKv4oP2FI'


def test_primer_Design_saving_to_text_file():

    files_ = [dedent(x) for x in
    ('''\
    >fw64 t-sequence
    atgactgctaacccttc
    >rv64 t-sequence
    catcgtaagtttcgaac''',

    '''\
    >fw64 t-sequence
    atgactgctaacccttcN
    >rv64 t-sequence
    catcgtaagtttcgaacN''',

    '''\
    >fw64 t-sequence
    atgactgctaacccttcN''',

    '''\
    >fw64 t-sequence
    atgactgctaacccttcN
    >rv64 t-sequence
    catcgtaagtttcgaac''',

    '''\
    >rv64 t-sequence
    catcgtaagtttcgaacN''',

    '''\
    >rv64 t-sequence
    catcgtaagtttcgaacN
    >fw64 t-sequence
    atgactgctaacccttc''',

    '''\
    >fw64 t-sequenceZ
    atgactgctaacccttc
    >rv64 t-sequenceZ
    catcgtaagtttcgaac''',

    '''\
    >fw64 t-sequenceZ
    atgactgctaacccttc
    >rv64 t-sequenceZ
    catcgtaagtttcgaac
    >fw64 t-sequence
    atgactgctaacccttc
    >rv64 t-sequence
    catcgtaagtttcgaac''',

    '''\
    >fw64x t-sequence
    atgactgctaacccttc
    >rv64x t-sequence
    catcgtaagtttcgaac''',

    '''\
    >fw64x t-sequence
    atgactgctaacccttc
    >rv64x t-sequence
    catcgtaagtttcgaac
    >fw64 t-sequence
    atgactgctaacccttc
    >rv64 t-sequence
    catcgtaagtttcgaac''',
    )]

    templ=Dseqrecord("atgactgctaacccttccttggtgttgaacaagatcgacgacatttcgttcgaaacttacgatg")

    templ.accession = "t-sequence"

    try:
        os.remove("PRIMERS.TXT")
    except OSError:
        pass

    pf, pr = cloning_primers(templ, path="PRIMERS.TXT")

    assert os.path.isfile("PRIMERS.TXT")

    with open("PRIMERS.TXT", "r") as f: text=f.read().strip()

    assert text == files_[0]

    with open("PRIMERS.TXT", "w") as f: text=f.write(files_[1])

    pf, pr = cloning_primers(templ, path="PRIMERS.TXT")

    with open("PRIMERS.TXT", "r") as f: text=f.read().strip()

    assert text == files_[1]

    with open("PRIMERS.TXT", "w") as f: text=f.write(files_[2])

    pf, pr = cloning_primers(templ, path="PRIMERS.TXT")

    with open("PRIMERS.TXT", "r") as f: text=f.read().strip()

    assert text == files_[3]

    with open("PRIMERS.TXT", "w") as f: text=f.write(files_[4])

    pf, pr = cloning_primers(templ, path="PRIMERS.TXT")

    with open("PRIMERS.TXT", "r") as f: text=f.read().strip()

    assert text == files_[5]

    with open("PRIMERS.TXT", "w") as f: text=f.write(files_[6])

    pf, pr = cloning_primers(templ, path="PRIMERS.TXT")

    with open("PRIMERS.TXT", "r") as f: text=f.read().strip()

    assert text == files_[7]

    with open("PRIMERS.TXT", "w") as f: text=f.write(files_[8])

    pf, pr = cloning_primers(templ, path="PRIMERS.TXT")

    with open("PRIMERS.TXT", "r") as f: text=f.read().strip()

    assert text == files_[9]

    with open("PRIMERS.TXT", "w") as f: text=f.write('')

    pf, pr = cloning_primers(templ, path="PRIMERS.TXT")

    f, r = parse("PRIMERS.TXT", ds=False)

    assert f.description == "fw64 t-sequence"
    assert r.description == "rv64 t-sequence"

if __name__ == '__main__':
    pytest.cmdline.main([__file__, "-v", "-s"])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
'''
_pretty tests
'''
import sys
import pytest
import pydna

def test_pretty():

    import io
    from Bio import SeqIO
    from Bio.Alphabet.IUPAC import IUPACAmbiguousDNA

    with open("pydna_read_test.txt", 'r', encoding="utf-8") as f:
        raw = f.read()

    handle = io.StringIO(raw)

    sr = SeqIO.read(handle, "genbank", alphabet=IUPACAmbiguousDNA())

    s = sr.format("gb").strip()

    pretty_label = pydna._pretty.pretty_string( s[:55]+"circular"+s[63:] )[559:578]

    fe=sr.features[0]
    label_from_sr = fe.qualifiers["label"][0]                                   
    
    assert raw[295:312] == '/label=2micron 2µ'

    assert label_from_sr == "2micron 2µ"

    assert s[559:578] == '/label="2micron 2µ"'

    assert pretty_label == '/label="2micron 2µ"'

if __name__ == '__main__':
    pytest.cmdline.main([__file__, "-v", "-s"])



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import pytest
import sys
import pydna

from Bio.SeqIO import read as BPread
from Bio.SeqIO import parse as BPparse

def pydna_read_test():
    print("sys.getdefaultencoding()",sys.getdefaultencoding())
    import locale
    print("locale.getpreferredencoding()", locale.getpreferredencoding())
    assert pydna.read("pydna_read_test.txt").format("gb")[559:578] == '/label="2micron 2µ"'

def test_parse_and_read_with_biopython_and_pydna():

    q = BPread("read1.gb", "gb")
    w = BPread("read2.gb", "gb")
    e = BPread("read3.fasta", "fasta")
    r = BPread("read4.fasta", "fasta")

    #a, b = BPparse("pth1.txt", "gb")
    with open("pth1.txt", "r", encoding="utf-8") as f:
        a, b = BPparse(f, "gb")
        
    print("|"+a.features[13].qualifiers['label'][0]+"|")
    print("|"+a.format("gb")[3314:3324]+"|")
    
    assert a.features[13].qualifiers['label'][0] == '2micron 2µ'
    assert a.format("gb")[3314:3324] == '2micron 2µ'

    x, y = pydna.parse("pth1.txt")

    assert "".join(a.format("gb").splitlines()[1:]) == "".join(x.format("gb").splitlines()[1:])
    assert "".join(b.format("gb").strip().splitlines()[4:]) == "".join(y.format("gb").splitlines()[4:])

def test_read_from_string():

    input_ ='''
            LOCUS       New_DNA                    4 bp ds-DNA     linear       30-MAR-2013
            DEFINITION  .
            ACCESSION
            VERSION
            SOURCE      .
              ORGANISM  .
            COMMENT
            COMMENT     ApEinfo:methylated:1
            FEATURES             Location/Qualifiers
                 misc_feature    2..3
                                 /label=NewFeature
                                 /ApEinfo_fwdcolor=cyan
                                 /ApEinfo_revcolor=green
                                 /ApEinfo_graphicformat=arrow_data {{0 1 2 0 0 -1} {} 0}
                                 width 5 offset 0
            ORIGIN
                    1 acgt
            //
            '''
    a = pydna.read(input_)
    
    assert str(a.seq)=="ACGT"

    input_ ='''>hej
               acgt'''
               
    assert str(a.seq)=="ACGT"

    input_ ='''
            LOCUS       New_DNA                    4 bp ds-DNA     linear       30-MAR-2013
            DEFINITION  .
            ACCESSION
            VERSION
            SOURCE      .
              ORGANISM  .
            COMMENT
            COMMENT     ApEinfo:methylated:1
            FEATURES             Location/Qualifiers
                 misc_feature    2..3
                                 /label=NewFeature
                                 /ApEinfo_fwdcolor=cyan
                                 /ApEinfo_revcolor=green
                                 /ApEinfo_graphicformat=arrow_data {{0 1 2 0 0 -1} {} 0}
                                 width 5 offset 0
            ORIGIN
                    1 acgt
            //
            '''
    a = pydna.read(input_)
    assert str(a.seq)=="ACGT"

    input_ ='''>hej
                acgt'''
    assert str(a.seq)=="ACGT"

    input_ ='''>hej öööh!
                acgt'''
    assert str(a.seq)=="ACGT"

    input_ ='''
                LOCUS       New_DNA                    4 bp ds-DNA     linear       30-MAR-2013
                DEFINITION  öööh!
                ACCESSION
                VERSION
                SOURCE      .
                  ORGANISM  .
                COMMENT
                COMMENT     ApEinfo:methylated:1
                FEATURES             Location/Qualifiers
                     misc_feature    2..3
                                     /label=öööh!
                                     /ApEinfo_fwdcolor=cyan
                                     /ApEinfo_revcolor=green
                                     /ApEinfo_graphicformat=arrow_data {{0 1 2 0 0 -1} {} 0}
                                     width 5 offset 0
                ORIGIN
                        1 acgt
                //
            '''
    a = pydna.read(input_)
    assert str(a.seq)=="ACGT"

def test_read_from_unicode():
    with open("pth1.txt", "r", encoding="utf-8") as f: 
        text = f.read()
    assert type(text) == str
    x,y = pydna.parse( text )
    assert x.format()[3314:3324] == '2micron 2µ'

def test_read_from_file():
    a = pydna.read("read1.gb")
    b = pydna.read("read2.gb")
    c = pydna.read("read3.fasta")
    d = pydna.read("read4.fasta")
    x,y = pydna.parse( "pth1.txt" )

    a.format("gb")
    b.format("gb")
    c.format("gb")
    d.format("gb")
    x.format("gb")
    y.format("gb")
    assert x.format()[3314:3324] == '2micron 2µ'
    assert x.features[13].qualifiers['label'][0] == u'2micron 2µ'
    assert str(a.seq).lower()==str(b.seq).lower()==str(c.seq).lower()==str(d.seq).lower()

if __name__ == '__main__':
    pytest.cmdline.main([__file__, "-v", "-s"])




#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys
import pytest

from pydna import read, parse, pcr, Assembly, eq


def test_MXblaster1():
    ''' test MXblaster1'''

    primer = parse("primers.fas", ds=False)
    primer     = primer[::-1]
    primer     = primer[37:]

    for i,p in enumerate(primer):
        assert int(p.id.split("_")[0]) == i

    ''' These are PCRs to get the genes and the terminator-promoters '''
    AgTEFp          = pcr(primer[524],primer[523],read("pAG25.gb"))
    hph             = pcr(primer[502],primer[501],read("pAG32.gb"))
    KlLEU2tt        = pcr(primer[520],primer[519],read("KlLEU2tt.gb"))

    ''' The Gal1 promoter-ISceI fragment is made in two steps '''
    gal1_ISceI_1    = pcr(primer[234],primer[316],read("pGSHU 7180bp.gb"))
    gal1_ISceI_2    = pcr(primer[562],primer[234],gal1_ISceI_1)
    AgTEFt          = pcr(primer[522],primer[521],read("pAG25.gb"))

    ''' load pCAPs and pCAPs-pSU0 sequences as Dseqrecord objects '''
    pCAPs      = read("pCAPs-AjiI.gb")
    pCAPs_pSU0 = read("pCAPs-pSU0.gb")

    # cut the pCAPs vectors for cloning
    from Bio.Restriction import EcoRV, ZraI
    pCAPs_ZraI              = pCAPs.cut(ZraI).pop()
    pCAPs_PCR_prod          = pcr(primer[492],primer[493], pCAPs)
    pCAPs_EcoRV             = pCAPs.cut(EcoRV).pop()
    pCAPs_pSU0_E_Z, stuffer = pCAPs_pSU0.cut(EcoRV, ZraI)


    # make the pCAPs clones, six altoghether
    pCAPs_ZraI_AgTEFp       = (pCAPs_ZraI + AgTEFp ).looped()

    pCAPs_PCR_prod_hph      = (pCAPs_PCR_prod  + hph ).looped()
    pCAPs_EcoRV_KlLEU2tt    = (pCAPs_EcoRV     + KlLEU2tt ).looped()

    pCAPs_ZraI_KlLEU2tt         = (pCAPs_ZraI      + KlLEU2tt ).looped()
    pCAPs_PCR_prod_gal1_ISceI_2 = (pCAPs_PCR_prod  + gal1_ISceI_2 ).looped()
    pCAPs_EcoRV_AgTEFt          = (pCAPs_EcoRV     + AgTEFt ).looped()

    # PCR each clone for the assembly in pCAPs

    A_AgTEFp_b     = pcr([primer[167],primer[493]], pCAPs_ZraI_AgTEFp)
    B_hph_c        = pcr([primer[467],primer[468]], pCAPs_PCR_prod_hph)
    C_KlLEU2tt_d   = pcr([primer[492],primer[166]], pCAPs_EcoRV_KlLEU2tt)

    #Homologous recombination of the two tp-gene-tp building blocks

    a=Assembly( ( A_AgTEFp_b,
                  B_hph_c,
                  C_KlLEU2tt_d,
                  pCAPs_pSU0_E_Z) , limit=28)

    YPK0_AgTEFp_hph_KlLEU2tt = a.circular_products[0]

    AgTEFp_hph_KlLEU2tt_2 = pcr(primer[166],primer[167], YPK0_AgTEFp_hph_KlLEU2tt)

    A_KlLEU2tt_b   = pcr([primer[167],primer[567]], pCAPs_ZraI_KlLEU2tt)
    B_gal1_ISceI_c = pcr([primer[467],primer[468]], pCAPs_PCR_prod_gal1_ISceI_2)
    C_AgTEFt_d     = pcr([primer[568],primer[166]], pCAPs_EcoRV_AgTEFt)

    a=Assembly(( A_KlLEU2tt_b,
                 B_gal1_ISceI_c,
                 C_AgTEFt_d,
                 pCAPs_pSU0_E_Z), limit=25)
    YPK0_KlLEU2tt_gal1_ISceI_AgTEFt = a.circular_products[0]

    KlLEU2tt_gal1_ISceI_AgTEFt_2 =  pcr(primer[166],primer[167], YPK0_KlLEU2tt_gal1_ISceI_AgTEFt)

    a=Assembly(( AgTEFp_hph_KlLEU2tt_2,
                 KlLEU2tt_gal1_ISceI_AgTEFt_2,
                 pCAPs_pSU0_E_Z), limit=61)

    pCAPs_MX4blaster1 = a.circular_products[0]

    pCAPs_MX4blaster1=pCAPs_MX4blaster1.synced("tcgcgcgtttcggtgatgacggtgaaaacc")

    assert pCAPs_MX4blaster1.seguid()=="X9WqaNk2lw6FbZlJr995MaDfn-M"

    from Bio.Restriction import AjiI, AgeI

    AX023560 = read("AX023560.gb")

    GAL10prom_slice= slice(AX023560.features[1].location.start,
                           AX023560.features[1].location.end)

    GAL10prom = AX023560[GAL10prom_slice]

    assert GAL10prom.seq == AX023560.features[1].extract(AX023560).seq

    GIN11M86 = read("GIN11M86.gb")

    GAL_GIN = pcr(primer[592],primer[593], GAL10prom + GIN11M86)

    assert GAL_GIN.seguid() == "7Lkfw8dsz9_kkBU3XXnz4KAON3A"

    assert pCAPs.seguid() =="-XHU8OxITyHGTl9XtMrJ4NvEv3o"

    pCAPs_GAL_GIN = ( pCAPs.cut(AjiI).pop() + GAL_GIN ).looped()

    assert pCAPs_GAL_GIN.seguid() == "T1eWCPIXPlq2HriSfpFSNnGwmd4"

    GAL_GIN2 = pcr(primer[592], primer[467], pCAPs_GAL_GIN)

    assert GAL_GIN2.seguid() =="zdIU4vjdfOxLkTTnKzIxhphnewg"

    assert pCAPs_MX4blaster1.seguid() =="X9WqaNk2lw6FbZlJr995MaDfn-M" # 9772bp__a

    pCAPs_MX4blaster1_AgeI = pCAPs_MX4blaster1.cut(AgeI).pop()

    pCAPs_MX4blaster1_AgeI.seq = pCAPs_MX4blaster1_AgeI.seq.fill_in()

    a=Assembly([GAL_GIN2, pCAPs_MX4blaster1_AgeI], limit=30)
       
    pCAPs_MX4blaster2 = a.circular_products[0]

    pCAPs_MX4blaster2 = pCAPs_MX4blaster2.synced("tcgcgcgtttcggtgatgacggtgaaaacc")
    
    assert len(pCAPs_MX4blaster2) == 10566 
    pCAPs_MX4blaster2_old = read("pMX4blaster2_old.gb")
    
    assert len(pCAPs_MX4blaster2_old) == 10566 
    assert pCAPs_MX4blaster2_old.seguid() == "7B4KKAeM2x8npjkp5U942rtMbB8"
    assert eq(pCAPs_MX4blaster2, pCAPs_MX4blaster2_old)
    assert pCAPs_MX4blaster2.seguid() == "7B4KKAeM2x8npjkp5U942rtMbB8"

    
if __name__ == '__main__':
    pytest.cmdline.main([__file__, "-v", "-s"])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
'''
test pGUP1
'''
import pytest
import pydna   

def test_assemble_pGUP1():
    ''' test pGUP1'''
   
    GUP1rec1sens = pydna.read("GUP1rec1sens.txt")
    GUP1rec2AS = pydna.read("GUP1rec2AS.txt")
    GUP1_locus = pydna.read("GUP1_locus.gb")
    pGREG505 = pydna.read("pGREG505.gb")

    insert = pydna.pcr(GUP1rec1sens, GUP1rec2AS, GUP1_locus)

    from Bio.Restriction import SalI

    lin_vect, his3 = pGREG505.cut(SalI)

    a = pydna.Assembly([insert, lin_vect], limit=28)
    
    pGUP1 = a.circular_products[0]
   
    pGUP1 = pGUP1.synced(pGREG505.seq[:50])    
    
    pGUP1_correct = pydna.read("pGUP1_correct.gb")        
    
    assert len(pGUP1_correct) == 9981
    assert len(pGUP1) == 9981
    assert pydna.eq(pGUP1, pGUP1_correct)     
    assert pGUP1_correct.seguid() == "42wIByERn2kSe_Exn405RYwhffU"       
    assert pGUP1.seguid() == "42wIByERn2kSe_Exn405RYwhffU"    

if __name__ == '__main__':
    pytest.cmdline.main([__file__, "-v", "-s"])


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright 2013 by Björn Johansson.  All rights reserved.
# This code is part of the Python-dna distribution and governed by its
# license.  Please see the LICENSE.txt file that should have been included
# as part of this package.

from .dseqrecord import Dseqrecord
import pathlib

class GenbankFile(Dseqrecord):

    def __init__(self, record, *args, path=None, **kwargs):
        super().__init__(record, *args, **kwargs)
        self.path=path

    def __repr__(self):
        '''returns a short string representation of the object'''
        return "File({})({}{})".format(self.id, {True:"-", False:"o"}[self.linear],len(self))
        
    def _repr_pretty_(self, p, cycle):
        '''returns a short string representation of the object'''
        p.text("File({})({}{})".format(self.id, {True:"-", False:"o"}[self.linear],len(self)))
            
    def _repr_html_(self):
        return "<a href='{path}' target='_blank'>{path}</a><br>".format(path=self.path)
        



#!/usr/bin/env python
# -*- coding: utf-8 -*-
'''
test cut
'''
import pytest
import sys

from pydna import pcr, cloning_primers, read, Dseqrecord

from Bio.Restriction import EcoRI

def test_cut_feat():
    puc19 = read('PUC19_MarkBudde.gb')
    pf, pr = cloning_primers(puc19)
    pcrProd = pcr(pf, pr, puc19)
    assert len(pcrProd.features) == 23
    #print len(pcrProd.cut(EcoRI)[1].features)
    assert len(pcrProd.cut(EcoRI)[1].features) == 17

    def amplicon_to_dseqrecord(a):
        d = Dseqrecord(a.seq)
        d.features = a.features
        return d

    pcrProdDseqrecord = amplicon_to_dseqrecord(pcrProd)
    assert len(pcrProdDseqrecord.cut(EcoRI)[1].features) == 17

if __name__ == '__main__':
    pytest.cmdline.main([__file__, "-v", "-s"])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
'''
test parse
'''

import pytest
import sys

from pydna import parse, read

def test_parse1():
    ''' test parsing fasta sequences from a text'''

    text   =  '''
            points....: 1

            The sequence seq below represents a double stranded linear DNA molecule.

            >seq
            CTCCCCTATCACCAGGGTACCGATAGCCACGAATCT

            Give the sequence(s) of the fragment(s) formed after digesting seq
            with the restriction enzyme Acc65I in the order that they appear in seq.

            Use FASTA format and give the Watson strand(s) in 5'-3' direction below.
            Give the sequences the names frag1,frag2,... etc.
            >frag1
            CTCCCCTATCACCAGG

            >frag2
            GTACCGATAGCCACGAATCT

            *********** Question 4 ***********

            QuestionID:
            '''
    result = parse(text)

    correct = ['CTCCCCTATCACCAGGGTACCGATAGCCACGAATCT',
               'CTCCCCTATCACCAGG',
               'GTACCGATAGCCACGAATCT']

    assert [str(s.seq) for s in result] == correct
    assert [s.linear for s in result] == [True,True,True]


    input =   '''
            LOCUS       ScCYC1                   330 bp    DNA              UNK 01-JAN-1980
            DEFINITION  ScCYC1
            ACCESSION   ScCYC1
            VERSION     ScCYC1
            KEYWORDS    .
            SOURCE      .
              ORGANISM  .
                        .
            FEATURES             Location/Qualifiers
            ORIGIN
                    1 ATGACTGAAT TCAAGGCCGG TTCTGCTAAG AAAGGTGCTA CACTTTTCAA GACTAGATGT
                   61 CTACAATGCC ACACCGTGGA AAAGGGTGGC CCACATAAGG TTGGTCCAAA CTTGCATGGT
                  121 ATCTTTGGCA GACACTCTGG TCAAGCTGAA GGGTATTCGT ACACAGATGC CAATATCAAG
                  181 AAAAACGTGT TGTGGGACGA AAATAACATG TCAGAGTACT TGACTAACCC AAAGAAATAT
                  241 ATTCCTGGTA CCAAGATGGC CTTTGGTGGG TTGAAGAAGG AAAAAGACAG AAACGACTTA
                  301 ATTACCTACT TGAAAAAAGC CTGTGAGTAA
            //
            '''
    result = parse(input).pop()

    assert str(result.seq) == str(read(input).seq)
    correct = '''ATGACTGAATTCAAGGCCGGTTCTGCTAAGAAAGGTGCTACACTTTTCAAGACTAGATGTCTACAATGCCACACCGTGGAAAAGGGTGGCCCACATAAGGTTGGTCCAAACTTGCATGGTATCTTTGGCAGACACTCTGGTCAAGCTGAAGGGTATTCGTACACAGATGCCAATATCAAGAAAAACGTGTTGTGGGACGAAAATAACATGTCAGAGTACTTGACTAACCCAAAGAAATATATTCCTGGTACCAAGATGGCCTTTGGTGGGTTGAAGAAGGAAAAAGACAGAAACGACTTAATTACCTACTTGAAAAAAGCCTGTGAGTAA'''

    assert str(result.seq) == correct
    assert result.linear   == True 
    assert result.circular == False 

    seqs = parse('RefDataBjorn.fas')

    assert len(seqs) == 771    
    assert list(set([len (a) for a in seqs])) == [901]
    pAG25 = read("pAG25.gb")

    assert pAG25.circular == True 
    assert pAG25.linear   == False

    pCAPs = read("pCAPs.gb")

    assert pCAPs.circular == True 
    assert pCAPs.linear   == False

    pUC19 = read("pUC19.gb")

    assert pUC19.circular == True 
    assert pUC19.linear   == False



def test_parse2():
    from Bio.Alphabet.IUPAC import IUPACAmbiguousDNA

    seqs = parse('RefDataBjorn.fas')

    assert len(seqs) == 771
    assert list(set([len (a) for a in seqs])) == [901]

    for i,s in enumerate(seqs):
        a = s.description
        b = a.split()
        c =  "|".join([b[0],b[1],b[3]])
        s.id = b[2].replace(" ","_")+"_"+str(i)
        s.description = ""
        if b[3]=="Zenion hololepis":
            s.id = b[3].replace(" ","_")+"_"+str(i)
        s.seq.alphabet = IUPACAmbiguousDNA()


if __name__ == '__main__':
    pytest.cmdline.main([__file__, "-v", "-s"])











#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os
import shutil
import tempfile
import pytest

def main():

    if os.getenv("CI"):
        ci = os.getenv("DRONE") or os.getenv("TRAVIS") or os.getenv("APPVEYOR")
        print("Tests run on continuous integration server {}".format(ci))
        cwd = os.getcwd()
        print("current working directroy:", cwd)
        
        os.environ["pydna_data_dir"]    = os.path.join(cwd,"DATA")
        os.environ["pydna_log_dir"]     = os.path.join(cwd,"DATA")
        os.environ["pydna_config_dir"]  = os.path.join(cwd,"DATA")

        # create data directory if not present
        try:
            os.makedirs( os.environ["pydna_data_dir"] )
        except OSError:
            if os.path.isdir( os.environ["pydna_data_dir"] ):
                pass
            else:
                raise

        print('os.environ["pydna_data_dir"] = ',  os.environ["pydna_data_dir"])
        print('os.environ["pydna_log_dir"] = ',   os.environ["pydna_log_dir"])
        print('os.environ["pydna_config_dir"] = ',os.environ["pydna_config_dir"])
        print('')

    else:
        print("Tests run locally")
        os.environ["pydna_data_dir"] = tempfile.mkdtemp(prefix="pydna_data_dir_test_")
        os.environ["pydna_log_dir"]  = tempfile.mkdtemp(prefix="test_pydna_log_dir_test")

    print("      _             _                     _               _            _               _ _       _ ")
    print("     | |           | |                   | |             | |          | |             (_) |     | |")
    print("  ___| |_ __ _ _ __| |_   _ __  _   _  __| |_ __   __ _  | |_ ___  ___| |_   ___ _   _ _| |_ ___| |")
    print(" / __| __/ _` | '__| __| | '_ \| | | |/ _` | '_ \ / _` | | __/ _ \/ __| __| / __| | | | | __/ _ \ |")
    print(" \__ \ || (_| | |  | |_  | |_) | |_| | (_| | | | | (_| | | ||  __/\__ \ |_  \__ \ |_| | | ||  __/_|")
    print(" |___/\__\__,_|_|   \__| | .__/ \__, |\__,_|_| |_|\__,_|  \__\___||___/\__| |___/\__,_|_|\__\___(_)")
    print("                         | |     __/ |                                                             ")
    print("                         |_|    |___/                                                              ")

    try:
        import coveralls
    except ImportError:
        print("python-coveralls NOT installed!")
        args = []
    else:
        del coveralls
        args = ["--cov=pydna", "--cov-report=html", "--cov-report=xml"]    
    try:
        import nbval
    except ImportError:
        print("nbval NOT installed!")
    else:
        del nbval
        args.append("--nbval")   

    args = [".", "-v", "-s"] + args 
    cwd = os.getcwd()
    os.chdir("tests")
    pytest.cmdline.main(args)
    os.chdir(cwd)
    try:
        shutil.copy(os.path.join("tests","coverage.xml"), "coverage.xml")
    except FileNotFoundError:
        pass
    args = ["pydna", "--doctest-modules", "-v", "-s"]
    pytest.cmdline.main(args)

    print("                  _               _            _               _ _          __ _       _     _              _ _ ")
    print("                 | |             | |          | |             (_) |        / _(_)     (_)   | |            | | |")
    print("  _ __  _   _  __| |_ __   __ _  | |_ ___  ___| |_   ___ _   _ _| |_ ___  | |_ _ _ __  _ ___| |__   ___  __| | |")
    print(" | '_ \| | | |/ _` | '_ \ / _` | | __/ _ \/ __| __| / __| | | | | __/ _ \ |  _| | '_ \| / __| '_ \ / _ \/ _` | |")
    print(" | |_) | |_| | (_| | | | | (_| | | ||  __/\__ \ |_  \__ \ |_| | | ||  __/ | | | | | | | \__ \ | | |  __/ (_| |_|")
    print(" | .__/ \__, |\__,_|_| |_|\__,_|  \__\___||___/\__| |___/\__,_|_|\__\___| |_| |_|_| |_|_|___/_| |_|\___|\__,_(_)")
    print(" | |     __/ | ")
    print(" |_|    |___/  ")

if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# doctest: +NORMALIZE_WHITESPACE
# doctest: +SKIP
'''This module provides a way to clean up broken Genbank files enough to pass the BioPython Genbank parser
This parser is based on pyparsing.
Almost all of this code was lifted from BioJSON (https://github.com/levskaya/BioJSON) by Anselm Levskaya.
The code put up was not accompanied by any software licence.

There are some modifications to deal with fringe cases.

The parser first produces JSON as an intermediate format which is then formatted back into a string in Genbank format.

The parser is not complete, so some fields do not survive the roundtrip (see below). This should not be a difficult fix, though.
The returned result has two properties,  .jseq which is the intermediate JSON produced by the parser and .gbtext 
which is the formatted genbank string.

'''

import os as _os
import re as _re
import json as _json
from pyparsing import *

from .dseqrecord import Dseqrecord

GoodLocus =  ( Literal("LOCUS") +
               Word(alphas+nums+'-_().'+'\\').setResultsName("name") +
               Word(nums).setResultsName("size")+Suppress(CaselessLiteral('bp')) +
               Word(alphas+'-').setResultsName("seqtype") +
               (CaselessLiteral("linear")|CaselessLiteral("circular")).setResultsName("topology") +
               Optional(Word(alphas), default="   ").setResultsName("divcode") +
               Regex(r"(\d{2})-(\S{3})-(\d{4})").setResultsName("date") )

# Older versions of ApE don't include a LOCUS name! Need separate def for this case:
BrokenLocus1 =( Literal("LOCUS").setResultsName("name") +
                Word(nums).setResultsName("size")+Suppress(CaselessLiteral('bp')) +
                Word(alphas+'-').setResultsName("seqtype") +
                (CaselessLiteral("linear")|CaselessLiteral("circular")).setResultsName("topology") +
                Optional(Word(alphas), default="   ").setResultsName("divcode") +
                Regex(r"(\d{2})-(\S{3})-(\d{4})").setResultsName("date") )

# LOCUS       YEplac181	5741 bp 	DNA	SYN
BrokenLocus2 =( Literal("LOCUS") +
                Word(alphas+nums+'-_().'+'\\').setResultsName("name") +
                Word(nums).setResultsName("size")+Suppress(CaselessLiteral('bp')) +
                Word(alphas+'-').setResultsName("seqtype") +
                Optional(CaselessLiteral("linear")|CaselessLiteral("circular"), default="linear").setResultsName("topology") +
                Optional(Word(alphas), default="   ").setResultsName("divcode") +
                Regex(r"(\d{2})-(\S{3})-(\d{4})").setResultsName("date") )

BrokenLocus3 =( Literal("LOCUS") +
                Word(alphas+nums+'-_().'+'\\').setResultsName("name") +
                Word(nums).setResultsName("size")+Suppress(CaselessLiteral('bp')) +
                Word(alphas+'-').setResultsName("seqtype") +
                Optional(CaselessLiteral("linear")|CaselessLiteral("circular"), default="linear").setResultsName("topology") +
                Word(alphas).setResultsName("divcode") +
                Optional(Regex(r"(\d{2})-(\S{3})-(\d{4})").setResultsName("date"), default="19-MAR-1970").setResultsName("date") )

LocusEntry = (GoodLocus|BrokenLocus1|BrokenLocus2|BrokenLocus3)

#===============================================================================
# Generic Entry

# this catches everything but the FEATURES and SEQUENCE entries, really should add parsing code for
# ACCESSION, COMMENTS, REFERENCE, ORGANISM, etc.
# (Though these entries are generally useless when it comes to hacking on DNA)

# All entries in a genbank file headed by an all-caps title with no space between start-of-line and title
CapWord = Word("ABCDEFGHIJKLMNOPQRSTUVWXYZ")
# after titled line, all subsequent lines have to have at least one space in front of them
# this is how we split up the genbank record
SpacedLine =  White(min=1) + CharsNotIn("\n") + LineEnd()
#HeaderLine = CapWord + CharsNotIn("\n") + LineEnd()
GenericEntry =  Group(CapWord + Combine(CharsNotIn("\n") + LineEnd() +\
                             ZeroOrMore( SpacedLine ))).setResultsName("generics",listAllMatches=True)


#===============================================================================
# Definition Entry
#SuppressedSpacedLine =  Suppress(White(min=1)) + CharsNotIn("\n") + LineEnd()
#DefinitionEntry =  Suppress(Literal("DEFINITION")) + Combine(CharsNotIn("\n") + LineEnd() + ZeroOrMore( SuppressedSpacedLine ))

#===============================================================================
# GenBank Feature Table Parser

#==== Genbank Location String Parser
#
# a string of slices w. functional modifiers that go at most two levels deep
# single position is just a number i.e. 23423
# slice is N1..N2  w. N1<N2
# i.e.
# 23..88  --> seq[23:89] in python syntax (genbank uses inclusive slicing)
# 234..555 
# complement(234..434) --> rc(seq[234:435])
# join(23..343,454..666,777..999) --> seq[23:344]+seq[454:667]+seq[777:1000]
# complement(join(23..343,454..666,777..999))
# join(complement(34..123),complement(333..565))
#
# additionally the slices can have ambiguous locs like <454..999 or 232..>331
# also note the dumb 34.38 fuzzy slice notation
# i.e. <45..900  says the real feature starts "somewhere" before pos 45
#       45.48 says feature somewhere between bases 45->48
# lot of weird annotations best avoided because they connote ~useless knowledge for synthetic design
#
# if you don't know where something is, don't use it or guess and move on

LPAREN = Suppress("(")
RPAREN = Suppress(")")
SEP = Suppress(Literal(".."))

#recognize numbers w. < & > uncertainty specs, then strip the <> chars to make it fixed
gbIndex = Word(nums+"<>").setParseAction(lambda s,l,t: int(t[0].replace("<","").replace(">","")) )
SimpleSlice=Group(gbIndex + SEP + gbIndex) | Group(gbIndex).setParseAction(lambda s,l,t: [[t[0][0],t[0][0]]])

#recursive def for nested function syntax:  f( g(), g() )
complexSlice = Forward()
complexSlice << (Literal("complement") | Literal("join")) + LPAREN + ( delimitedList(complexSlice) | delimitedList(SimpleSlice) ) + RPAREN 
featLocation = Group( SimpleSlice | complexSlice)

def parseGBLoc(s,l,t):
    """retwingles parsed genbank location strings, assumes no joins of RC and FWD sequences """
    strand = 1
    locationlist=[]
    
    #see if there are any complement operators
    for entry in t[0]:
        if entry=="complement": strand=-1

    for entry in t[0]:
        if type(entry)!=type("string"):
            locationlist.append([entry[0],entry[1]])
            
    #return locationlist and strand spec
    return [ ['location', locationlist ], ['strand',strand] ]

featLocation.setParseAction(parseGBLoc)

#==== Genbank Feature Key-Value Pairs
def strip_multiline(s,l,t):
    whitespace=_re.compile("[\n]{1}[ ]+")
    return whitespace.sub(" ",t[0])
def toInt(s,l,t):
    return int(t[0])

# Quoted KeyVal:   /key="value"
QuoteFeaturekeyval = Group(Suppress('/') + Word(alphas+nums+"_-") + Suppress('=') + QuotedString('"',multiline=True).setParseAction( strip_multiline ) )

# UnQuoted KeyVal: /key=value  (I'm assuming it doesn't do multilines this way? wrong! ApE does store long labels this way! sigh.)
#NoQuoteFeaturekeyval = Group(Suppress('/') + Word(alphas+nums+"_-") + Suppress('=') + OneOrMore(CharsNotIn("\n")) )
keyvalspacedline = White(exact=21)+CharsNotIn("/")+OneOrMore(CharsNotIn("\n"))+LineEnd()
NoQuoteFeaturekeyval = Group(Suppress('/') + Word(alphas+nums+"_-") + Suppress('=') + Combine(CharsNotIn("\n") + LineEnd() + ZeroOrMore( keyvalspacedline )))

# Special Case for Numerical Vals:  /bases=12  OR  /bases="12"
NumFeaturekeyval = Group( Suppress('/') + Word(alphas+nums+"_-") + Suppress('=') +\
                         (Suppress("\"") + Word(nums).setParseAction(toInt) + Suppress("\"") ) | \
                         (Word(nums).setParseAction(toInt) ) \
                         )

# Key Only KeyVal: /pseudo
# post-parse convert it into a pair to resemble the structure of the first three cases i.e. [pseudo, True]
FlagFeaturekeyval = Group(Suppress('/') + Word(alphas+nums+"_-")).setParseAction(lambda s,l,t: [[t[0][0],True]] )

Feature = Group( Word(alphas+nums+"_-").setParseAction(lambda s,l,t: [ ["type", t[0]] ] ) +\
                 featLocation.setResultsName("location") +\
                 OneOrMore( NumFeaturekeyval | QuoteFeaturekeyval | NoQuoteFeaturekeyval | FlagFeaturekeyval ) \
                 )

FeaturesEntry = Literal("FEATURES") + Literal("Location/Qualifiers") + Group(OneOrMore(Feature)).setResultsName("features")

#===============================================================================
# GenBank Sequence Parser

# sequence is just a column-spaced big table of dna nucleotides
# should it recognize full IUPAC alphabet?  NCBI uses n for unknown region
Sequence = OneOrMore(Suppress(Word(nums)) + OneOrMore(Word("ACGTacgtNn")))

# Group(  ) hides the setResultsName names def'd inside, such that one needs to first access this group and then access the dict of contents inside
SequenceEntry = Suppress(Literal("ORIGIN")) + Sequence.setParseAction(lambda s,l,t: "".join(t) ).setResultsName("sequence")


#===============================================================================
# Final GenBank Parser

#GB files with multiple records split by "//" sequence at beginning of line
GBEnd = Literal("//")

#Begin w. LOCUS, slurp all entries, then stop at the end!
GB = LocusEntry + OneOrMore(FeaturesEntry | SequenceEntry | GenericEntry) + GBEnd

#NCBI often returns sets of GB files
multipleGB = OneOrMore(Group(GB))

#===============================================================================
# End Genbank Parser
#===============================================================================


#===============================================================================
# Main JSON Conversion Routine

def strip_indent(str):
    whitespace=_re.compile("[\n]{1}(COMMENT){0,1}[ ]+")
    return whitespace.sub("\n",str)

def concat_dict(dlist):
    """more or less dict(list of string pairs) but merges
    vals with the same keys so no duplicates occur
    """
    newdict={}
    for e in dlist:
        if e[0] in newdict.keys():
            newdict[e[0]]=(newdict[e[0]]+strip_indent(e[1]))
        else:
            newdict[e[0]]=strip_indent(e[1])
    return newdict

def toJSON(gbkstring):
    
    parsed = multipleGB.parseString(gbkstring)
    
    jseqlist=[]

    for seq in parsed:
        
        #for item in seq.asList():
        #    print(item)
        
        
        #import sys;sys.exit(42) 
        
        #Print to STDOUT some details (useful for long multi-record parses)
        #print(seq['name'], ":  length:", len(seq['sequence']) , " #features:" , len(seq['features'].asList()))
        
        #build JSON object
        
        nl=[]
        
        for a in list(map(dict, seq['features'].asList())):
            dct={}
            for key in a:
                val=a[key]
                #print(key, a[key])
                dct[key]=a[key]
                if isinstance(val, str):
                    dct[key]=a[key].strip()
            nl.append(dct)
            
        #import sys;sys.exit(42)    
        
        #print(list(map(dict, hej))[2]["codon_start"])

        jseq = { "__format__" : "jseq v0.1",
                 "name"       : seq["name"],
                 "size"       : seq["size"],
                 "seqtype"    : seq["seqtype"],
                 "divcode"    : seq["divcode"],
                 "date"       : seq["date"],
                 "topology"   : seq["topology"],
                 "sequence"   : seq["sequence"],
                 "features"   : nl,
                 "annotations" : concat_dict(seq['generics'])
                 }
        jseqlist.append(jseq)

    return jseqlist

def wrapstring(str_, rowstart, rowend, padfirst=True):
    """
    wraps the provided string in lines of length rowend-rowstart
    and padded on the left by rowstart.
    -> if padfirst is false the first line is not padded
    """
    rowlen  = rowend-rowstart
    leftpad = rowstart
    wrappedstr=""

    #no wrapping needed, single line
    if len(str_)/rowlen < 1:
        if padfirst:
            return leftpad*" "+str_+"\n"
        else:
            return str_+"\n"
        
    #multiple lines so wrap:
    for linenum in range(1+int(len(str_)/rowlen)):
        if linenum==0 and not padfirst:
            wrappedstr+= str[linenum*rowlen:(linenum+1)*rowlen]+"\n"
        else:
            wrappedstr+=" "*leftpad + str_[linenum*rowlen:(linenum+1)*rowlen]+"\n"
#    if str_.startswith("/translation="):
#        print(str_)
#        print(wrappedstr)
#        print(".................................")
    return wrappedstr

def locstr(locs,strand):
    "genbank formatted location string, assumes no join'd combo of rev and fwd seqs"
    # slice format is like: 1..10,20..30,101..200
    locstr=",".join(map((lambda x: str(x[0])+".."+str(x[1])), locs))
    if len(locs)>1:
        locstr=("join("+locstr+")")
    if int(strand)==-1:
        locstr=("complement("+locstr+")")
    return locstr

def originstr(sequence):
    "formats dna sequence as broken, numbered lines ala genbank"
    wordlen=10
    cols=6
    rowlen=wordlen*cols
    outstr=""
    for linenum in range( int(len(sequence)/rowlen)+1):
        pos=linenum*rowlen
        #position of string for this row, then six blocks of dna
        outstr+=(" "*9+str(pos+1))[-9:] + " " + \
           sequence[pos:pos+10]+" "+\
           sequence[pos+10:pos+20]+" "+\
           sequence[pos+20:pos+30]+" "+\
           sequence[pos+30:pos+40]+" "+\
           sequence[pos+40:pos+50]+" "+\
           sequence[pos+50:pos+60]+"\n"
    return outstr

def toGB(jseqs):
    "parses json jseq data and prints out ApE compatible genbank"

    #take first jseq from parsed list
    if type(jseqs)==type([]):
        jseq=jseqs[0]
    else:
        jseq=jseqs
    
    #construct the LOCUS header string
    #  LOCUS format: 
    #    Positions  Contents
    #    ---------  --------
    #    00:06      LOCUS
    #    06:12      spaces
    #    12:??      Locus name   
    #    ??:??      space
    #    ??:40      Length of sequence, right-justified
    #    40:44      space, bp, space
    #    44:47      Blank, ss-, ds-, ms-
    #    47:54      Blank, DNA, RNA, tRNA, mRNA, uRNA, snRNA, cDNA
    #    54:55      space
    #    55:63      Blank (implies linear), linear or circular
    #    63:64      space
    #    64:67      The division code (e.g. BCT, VRL, INV)
    #    67:68      space
    #    68:79      Date, in the form dd-MMM-yyyy (e.g., 15-MAR-1991)

              
    name    = jseq["name"] or "default"
    size    = jseq["size"] or "100"
    seqtype = jseq["seqtype"] or "DNA"
    prefix = ""
    for p in ["ds-", "ss-", "ms-"]:
        a, *b = seqtype.split(p)
        if b: 
            prefix=p
            seqtype=b.pop()
            break
    prefix  = prefix or "ds-"    
    topology = jseq["topology"] or "linear"
    divcode  = jseq["divcode"] or "   "
    date = jseq["date"] or "19-MAR-1970"

    locusstr = "LOCUS       {name:<24}{size:>4} bp {prefix}{seqtype:<4}    {topology:<8} {divcode} {date}\n".format(name=name, 
                                                                                                                    size=size,
                                                                                                                    prefix=prefix,
                                                                                                                    seqtype=seqtype, 
                                                                                                                    topology=topology,
                                                                                                                    divcode=divcode,
                                                                                                                    date=date)
                                        
    # All these fields are left empty
    gbprops="DEFINITION  .\n"+\
              "ACCESSION   \n"+\
              "VERSION     \n"+\
              "SOURCE      .\n"+\
              "ORGANISM  .\n"+\
              "COMMENT     \n"+\
              "COMMENT     ApEinfo:methylated:1\n"+\
              "FEATURES             Location/Qualifiers\n"

    #build the feature table
    featuresstr=""
    for feat in jseq["features"]:
        fstr= " "*5 + feat["type"] +\
              " "*(16-len(feat["type"])) + \
              wrapstring(locstr(feat["location"],feat["strand"]),21,80,False)
        for k in feat.keys():
            if k not in ["type","location","strand"]:
                #ApE idiosyncrasy: don't wrap val in quotation marks
                if k in ["ApEinfo_label","ApEinfo_fwdcolor","ApEinfo_revcolor","label"]:
                    fstr+=wrapstring("/"+str(k)+"="+str(feat[k]),21,80)
                #standard: wrap val in quotes
                else:
                    fstr+=wrapstring("/"+str(k)+"="+"\""+str(feat[k])+"\"",21,80)
        featuresstr+=fstr

    #the spaced, numbered sequence
    gborigin="ORIGIN\n"+\
              originstr(jseq["sequence"])+\
              "//\n"
    
    return locusstr+gbprops+featuresstr+gborigin
            
def gbtext_clean(gbtext):
    jseqlist=toJSON(gbtext)
    jseq = jseqlist.pop()
    class Result(object):
        pass
    result=Result()
    result.jseq   = jseq
    result.gbtext = toGB(jseq)
    return result
        
        
if __name__=="__main__":
    import os
    cache = os.getenv("pydna_cache")
    os.environ["pydna_cache"]="nocache"
    import doctest
    doctest.testmod()
    os.environ["pydna_cache"]=cache




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
'''Provides a class for downloading sequences from genbank.
'''
import pickle
import shelve
import os
import requests
import textwrap

def download_text(url):
    cached  = False
    refresh = False
    cache = shelve.open(os.path.join(os.environ["pydna_data_dir"], "web"), protocol=pickle.HIGHEST_PROTOCOL, writeback=False)
    key = str(url)

    if os.environ["pydna_cache"] in ("compare", "cached"):
        try:
            cached = cache[key]
        except KeyError:
            if os.environ["pydna_cache"] == "compare":
                raise Exception("no result for this key!")
            else:
                refresh = True

    if refresh or os.environ["pydna_cache"] in ("compare", "refresh", "nocache"):
        result = requests.get(url).text
    if os.environ["pydna_cache"] == "compare":
        if result!=cached:
            module_logger.warning('download error')

    if refresh or os.environ["pydna_cache"] == "refresh":
        cache = shelve.open(os.path.join(os.environ["pydna_data_dir"],"genbank"), protocol=pickle.HIGHEST_PROTOCOL, writeback=False)
        cache[key] = result

    elif cached and os.environ["pydna_cache"] not in ("nocache", "refresh"):
        result = cached

    cache.close()
    
    result = textwrap.dedent(result).strip()
    result = result.replace( '\r\n', '\n')
    result = result.replace( '\r',   '\n')
    return result

if __name__=="__main__":
    import os
    cache = os.getenv("pydna_cache")
    os.environ["pydna_cache"]="nocache"
    import doctest
    doctest.testmod()
    os.environ["pydna_cache"]=cache



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# doctest: +NORMALIZE_WHITESPACE
# doctest: +SKIP
'''This module provides functions for PCR. Primers with 5' tails as well as inverse PCR on
circular templates are handled correctly.

'''

import pickle
import shelve

import itertools

import re
import copy
import operator
import os

import logging
module_logger = logging.getLogger("pydna."+__name__)

from Bio.Seq                        import Seq
from Bio.SeqUtils.CheckSum          import seguid
from Bio.SeqFeature                 import SeqFeature
from Bio.SeqFeature                 import CompoundLocation
from Bio.SeqFeature                 import FeatureLocation
from .utils                         import rc
from .dseqrecord                    import Dseqrecord
from .primer                        import Primer
from .amplicon                      import Amplicon


def _annealing_positions(primer, template, limit=15):
    '''Finds the annealing position(s) for a primer on a template where the
    primer anneals perfectly with at least limit nucleotides in the 3' part.

    start is a position (integer)
    footprint1 and tail1 are strings.

    ::

        <- - - - - - - - - - template - - - - - - - - - - - - - >

        <------------- start ---->
     5'-...gctactacacacgtactgactgcctccaagatagagtcagtaaccacactcgat...3'
           ||||||||||||||||||||||||||||||||||||||||||||||||
                                  3'-gttctatctcagtcattggtgtATAGTG-5'

                                                        <tail>
                                     <---footprint----->
                                     <--------- primer ------>

    Parameters
    ----------
    primer : string
        The primer sequence 5'-3'

    template : string
        The template sequence 5'-3'

    limit : int = 15, optional
        footprint needs to be at least of length limit.

    Returns
    -------
    describe : list of tuples (int, string, string)
        [ (start1, footprint1, tail1), (start2, footprint2, tail2),..., ]
    '''

    if len(primer)<limit:
        return []
    prc = rc(primer)
    head = prc[:limit].upper()

    table = {"R":"(A|G)",
             "Y":"(C|T)",
             "S":"(G|C)",
             "W":"(A|T)",
             "K":"(G|T)",
             "M":"(A|C)",
             "B":"(C|G|T)",
             "D":"(A|G|T)",
             "H":"(A|C|T)",
             "V":"(A|C|G)",
             "N":"(A|G|C|T)"}

    for key in table:
        head=head.replace(key, table[key])

    positions = [m.start() for m in re.finditer('(?={})'.format(head), template, re.I)]

    if positions:
        tail = prc[limit:]
        length = len(tail)
        results = []
        for match_start in positions:
            tm = template[match_start+limit:match_start+limit+length]
            footprint = rc(template[match_start:match_start+limit]+"".join([b for a,b in itertools.takewhile(lambda x: x[0].lower()==x[1].lower(), list(zip(tail, tm)))]))
            results.append((match_start, footprint, primer[: len(primer) - len(footprint) ]))
        return results
    return []

class Anneal(object):
    '''

    Parameters
    ----------
    primers : iterable containing SeqRecord objects
        Primer sequences 5'-3'.

    template : Dseqrecord object
        The template sequence 5'-3'.

    limit : int, optional
        limit length of the annealing part of the primers.

    fprimerc : float, optional
        Concentration of forward primer in nM, set to 1000.0 nM by default

    rprimerc : float, optional
        Concentration of reverse primer in nM, set to 1000.0 nM by default

    saltc  : float, optional
        Salt concentration (monovalet cations) :mod:`tmbresluc` set to 50.0 mM by default

    Attributes
    ----------
    products: list
        A list of Amplicon objects, one for each primer pair that may form a PCR product.


    Examples
    --------
    >>> import pydna
    >>> template = pydna.Dseqrecord("tacactcaccgtctatcattatctactatcgactgtatcatctgatagcac")
    >>> from Bio.SeqRecord import SeqRecord
    >>> p1 = pydna.read(">p1\\ntacactcaccgtctatcattatc", ds = False)
    >>> p2 = pydna.read(">p2\\ngtgctatcagatgatacagtcg", ds = False)
    >>> ann = pydna.Anneal((p1, p2), template)
    >>> print(ann.report())
    Template name? 51 nt linear:
    Primer p1 anneals forward at position 23
    <BLANKLINE>
    Primer p2 anneals reverse at position 29
    >>> ann.products
    [Amplicon(51)]
    >>> amplicon_list = ann.products
    >>> amplicon = amplicon_list.pop()
    >>> amplicon
    Amplicon(51)
    >>> print(amplicon.figure())
    5tacactcaccgtctatcattatc...cgactgtatcatctgatagcac3
                               |||||||||||||||||||||| tm 55.9 (dbd) 60.5
                              3gctgacatagtagactatcgtg5
    5tacactcaccgtctatcattatc3
     ||||||||||||||||||||||| tm 54.6 (dbd) 58.8
    3atgtgagtggcagatagtaatag...gctgacatagtagactatcgtg5
    >>> amplicon.annotations['date'] = '02-FEB-2013'   # Set the date for this example to pass the doctest
    >>> print(amplicon)
    Dseqrecord
    circular: False
    size: 51
    ID: 51bp U96-TO06Y6pFs74SQx8M1IVTBiY
    Name: 51bp_PCR_prod
    Description: Product_p1_p2
    Number of features: 4
    /date=02-FEB-2013
    Dseq(-51)
    taca..gcac
    atgt..cgtg
    >>> print(amplicon.program())
    <BLANKLINE>
    Taq (rate 30 nt/s) 35 cycles             |51bp
    95.0°C    |95.0°C                 |      |Tm formula: Biopython Tm_NN
    |_________|_____          72.0°C  |72.0°C|SaltC 50mM
    | 03min00s|30s  \         ________|______|Primer1C 1.0µM
    |         |      \ 45.4°C/ 0min 2s| 5min |Primer2C 1.0µM
    |         |       \_____/         |      |GC 39%
    |         |         30s           |      |4-12°C

    >>>

    '''
    def __init__( self,
                  primers,
                  template,
                  limit=13,
                  primerc=1000.0, # nM
                  saltc=50):

        refresh = False
        cached  = None

        self.primerc=primerc
        self.saltc = saltc

        primers = [p for p in primers if p.seq]

        key = str(template.seguid()) + "|".join(sorted([seguid(p.seq) for p in primers]))+str(limit)

        if os.environ["pydna_cache"] in ("compare", "cached"):
            cache = shelve.open(os.path.join(os.environ["pydna_data_dir"], "amplify"), protocol=pickle.HIGHEST_PROTOCOL, writeback=False)
            try:
                cached = cache[key]
            except:
                if os.environ["pydna_cache"] == "compare":
                    raise Exception("no result for this key!")
                else:
                    refresh = True
            cache.close()

        if refresh or os.environ["pydna_cache"] in ("compare", "refresh", "nocache"):

            self.key = key
            self.template = copy.deepcopy(template)

            self.limit = limit
            self._products = None

            self.fwd_primers = []
            self.rev_primers = []

            twl = len(self.template.seq.watson)
            tcl = len(self.template.seq.crick)

            if self.template.linear:
                tw = self.template.seq.watson
                tc = self.template.seq.crick
            else:
                tw = self.template.seq.watson+self.template.seq.watson
                tc = self.template.seq.crick +self.template.seq.crick

            for p in primers:
                self.fwd_primers.extend((Primer(p,
                                                position = tcl-pos - min(self.template.seq.ovhg, 0),
                                                footprint = Seq(fp),
                                                tail = Seq(tl))
                                        for pos, fp, tl in _annealing_positions(
                                                            str(p.seq),
                                                            tc,
                                                            limit) if pos<tcl))
                self.rev_primers.extend((Primer(p,
                                                position = pos + max(0, self.template.seq.ovhg),
                                                footprint = Seq(fp),
                                                tail = Seq(tl))
                                         for pos, fp, tl in _annealing_positions(
                                                                         str(p.seq),
                                                                         tw,
                                                                         limit) if pos<twl))
            self.fwd_primers.sort(key = operator.attrgetter('position'))
            self.rev_primers.sort(key = operator.attrgetter('position'), reverse=True)

            for fp in self.fwd_primers:
                if fp.position-len(fp.footprint)>=0:
                    start = fp.position-len(fp.footprint)
                    end   = fp.position
                    self.template.features.append(SeqFeature(FeatureLocation(start, end),
                                                        type ="primer_bind",
                                                        strand = 1,
                                                        qualifiers = {"note":[fp.name],
                                                                      "ApEinfo_fwdcolor":["green"],
                                                                      "ApEinfo_revcolor":["red"]}))
                else:
                    start = len(self.template)-len(fp.footprint)+fp.position
                    end = start+len(fp.footprint)-len(self.template)
                    sf=SeqFeature(CompoundLocation([FeatureLocation(start,len(self.template)),
                                                    FeatureLocation(0, end)]),
                                                    type="primer_bind",
                                                    location_operator="join",
                                                    qualifiers = {"note":[fp.name]})

                    self.template.features.append(sf)

            for rp in self.rev_primers:
                if rp.position+len(rp.footprint)<=len(self.template):
                    start = rp.position
                    end   = rp.position + len(rp.footprint)
                    self.template.features.append(SeqFeature(FeatureLocation(start,end),
                                                        type ="primer_bind",
                                                        strand = -1,
                                                        qualifiers = {"note":[rp.name],
                                                                      "ApEinfo_fwdcolor":["green"],
                                                                      "ApEinfo_revcolor":["red"]}))
                else:
                    start = rp.position
                    end = rp.position+len(rp.footprint)-len(self.template)
                    self.template.features.append(SeqFeature(CompoundLocation([FeatureLocation(start,len(self.template)),
                                                                               FeatureLocation(0,end)]),
                                                        type ="primer_bind",
                                                        location_operator= "join",
                                                        strand = -1,
                                                        qualifiers = {"note":[rp.name]}))
            self.forward_primers = self.fwd_primers
            self.reverse_primers = self.rev_primers

        if os.environ["pydna_cache"] == "compare":
            self._compare(cached)

        if refresh or os.environ["pydna_cache"] == "refresh":
            self._save()

        elif cached and os.environ["pydna_cache"] not in ("nocache","refresh"):
            for key, value in list(cached.__dict__.items()):
                setattr(self, key, value )
            cache.close()

    def _compare(self, cached):
        if str(self) != str(cached):
            module_logger.warning('amplify error')

    def _save(self):
        cache = shelve.open(os.path.join(os.environ["pydna_data_dir"], "amplify"), protocol=pickle.HIGHEST_PROTOCOL, writeback=False)
        cache[self.key] = self
        cache.close()

    @property
    def products(self):

        if self._products:
            return self._products

        self._products = []

        for fp in self.fwd_primers:
            for rp in self.rev_primers:

                if self.template.circular and fp.position>rp.position:
                    tmpl = self.template.shifted(fp.position-len(fp.footprint))
                    tmpl = tmpl._multiply_circular(2)
                    tmpl = tmpl[:len(self.template) - (fp.position - rp.position) + len(rp.footprint) + len(fp.footprint)]

                elif self.template.circular:
                    tmpl = self.template._multiply_circular(3)
                    tmpl = tmpl[fp.position-len(fp.footprint)+len(self.template):rp.position+len(rp.footprint)+len(self.template)]
                else:
                    tmpl = self.template[fp.position-len(fp.footprint):rp.position+len(rp.footprint)]

                prd = ( Dseqrecord(fp.tail) + tmpl + Dseqrecord(rp.tail).reverse_complement())

                prd.add_feature( 0, len(fp), label=fp.id)
                prd.add_feature( len(prd)-len(rp),len(prd),label=rp.id, strand=-1)

                prd.name = "{0}bp_PCR_prod".format(len(prd))[:16]
                prd.id = "{0}bp {1}".format( str(len(prd))[:14], prd.seguid() )
                prd.description="Product_{0}_{1}".format( fp.description,
                                                          rp.description)

                self._products.append( Amplicon(prd,
                                                template=tmpl,
                                                forward_primer=fp,
                                                reverse_primer=rp,
                                                saltc=self.saltc,
                                                fprimerc=self.primerc,
                                                rprimerc=self.primerc))
                assert " " not in str(prd.seq.watson)
                assert " " not in str(prd.seq.crick)

        return self._products



    def report(self):
        '''This method is an alias of str(Annealobj).
        Returns a short string representation.
        '''
        return self.__str__()

    def __repr__(self):
        ''' returns a short string representation '''
        return "Reaction(products = {})".format(len(self.fwd_primers*len(self.rev_primers)))

    def __str__(self):
        '''returns a short report describing if or where primer
       anneal on the template.
       '''
        mystring = "Template {name} {size} nt {top}:\n".format(name=self.template.name,
                                                               size=len(self.template),
                                                               top={True:"circular",
                                                                    False:"linear"}[self.template.circular]
                                                                    )
        if self.fwd_primers:
            for p in self.fwd_primers:
                mystring += "Primer {name} anneals forward at position {pos}\n".format(name=p.name, pos=p.position)
        else:
            mystring += "No forward primers anneal...\n"
        mystring +="\n"
        if self.rev_primers:
            for p in self.rev_primers:
                mystring += "Primer {name} anneals reverse at position {pos}\n".format(name=p.name, pos=p.position)
        else:
             mystring += "No reverse primers anneal...\n"
        return mystring.strip()






def pcr(*args,  **kwargs):
    '''pcr is a convenience function for Anneal to simplify its usage,
    especially from the command line. If more than one PCR product is
    formed, an exception is raised.

    args is any iterable of sequences or an iterable of iterables of sequences.
    args will be greedily flattened.

    Parameters
    ----------

    args : iterable containing sequence objects
        Several arguments are also accepted.

    limit : int = 13, optional
        limit length of the annealing part of the primers.

    Notes
    -----

    sequences in args could be of type:

    * string
    * Seq
    * SeqRecord
    * Dseqrecord

    The last sequence will be interpreted as the template
    all preceeding sequences as primers.

    This is a powerful function, use with care!

    Returns
    -------

    product : Dseqrecord
        a Dseqrecord object representing the PCR product.
        The direction of the PCR product will be the same as
        for the template sequence.

    Examples
    --------

    >>> import pydna
    >>> template = pydna.Dseqrecord("tacactcaccgtctatcattatctactatcgactgtatcatctgatagcac")
    >>> from Bio.SeqRecord import SeqRecord
    >>> p1 = pydna.read(">p1\\ntacactcaccgtctatcattatc", ds = False)
    >>> p2 = pydna.read(">p2\\ncgactgtatcatctgatagcac", ds = False).reverse_complement()
    >>> pydna.pcr(p1, p2, template)
    Amplicon(51)
    >>> pydna.pcr([p1, p2], template)
    Amplicon(51)
    >>> pydna.pcr((p1,p2,), template)
    Amplicon(51)
    >>>

    '''

    from Bio.SeqRecord import SeqRecord
    # flatten args
    output = []
    stack = []
    stack.extend(reversed(args))
    while stack:
        top = stack.pop()
        if hasattr(top, "__iter__") and not isinstance(top, SeqRecord):
            stack.extend(reversed(top))
        else:
            output.append(top)
    new=[]

    for s in output:
        if isinstance(s, Seq):
            s = SeqRecord(s)
        elif isinstance(s, SeqRecord):
            pass
        elif hasattr(s, "watson"):
            s=s.watson
        elif isinstance(s, str):
            s = SeqRecord(Seq(s))
        else:
            raise TypeError("the record property needs to be a string, Seq, SeqRecord or Dseqrecord object")
        new.append(s)

    anneal_primers = Anneal(  new[:-1],
                              new[-1],
                              **kwargs)

    if anneal_primers:
        if len(anneal_primers.products) == 1:
            return anneal_primers.products[0]
        elif len(anneal_primers.products) == 0:
            raise Exception("No PCR products! {}".format(anneal_primers.report()))
        else:
            raise Exception("PCR not specific! {}".format(anneal_primers.report()))
    else:
        raise Exception(anneal_primers.report())
    return


def nopcr(*args,  **kwargs):
    '''no pcr is a convenience function for Anneal to simplify its usage,
    especially from the command line. If one or more PCR products are
    formed, an exception is raised.

    args is any iterable of sequences or an iterable of iterables of sequences.
    args will be greedily flattened.

    Parameters
    ----------

    args : iterable containing sequence objects
        Several arguments are also accepted.

    limit : int = 13, optional
        limit length of the annealing part of the primers.

    Notes
    -----

    sequences in args could be of type:

    * string
    * Seq
    * SeqRecord
    * Dseqrecord

    The last sequence will be interpreted as the template
    all preceeding sequences as primers.

    This is a powerful function, use with care!

    Returns
    -------

    product : Dseqrecord
        a Dseqrecord object representing the PCR product.
        The direction of the PCR product will be the same as
        for the template sequence.

    Examples
    --------

    >>> import pydna
    >>> template = pydna.Dseqrecord("tacactcaccgtctatcattatctactatcgactgtatcatctgatagcac")
    >>> from Bio.SeqRecord import SeqRecord
    >>> from Bio.Seq import Seq
    >>> p1 = SeqRecord(Seq("tacactcaccgtctatcattatc"))
    >>> p2 = SeqRecord(Seq("gtgctatcagatgatacagtG")) # This primer does not anneal
    >>> pydna.nopcr(p1, p2, template)
    True
    '''

    from Bio.SeqRecord import SeqRecord
    # flatten args
    output = []
    stack = []
    stack.extend(reversed(args))
    while stack:
        top = stack.pop()
        if hasattr(top, "__iter__") and not isinstance(top, SeqRecord):
            stack.extend(reversed(top))
        else:
            output.append(top)
    new=[]

    for s in output:
        if isinstance(s, Seq):
            s = SeqRecord(s)
        elif isinstance(s, SeqRecord):
            pass
        elif hasattr(s, "watson"):
            s=s.watson
        elif isinstance(s, str):
            s = SeqRecord(Seq(s))
        else:
            raise TypeError("the record property needs to be a string, a Seq object or a SeqRecord object")
        new.append(s)

    anneal_primers = Anneal(  new[:-1],
                              new[-1],
                              **kwargs)

    if anneal_primers:
        if len(anneal_primers.products) != 0:
            raise Exception("PCR products formed! {}".format(anneal_primers.report()))
    else:
        raise Exception(anneal_primers.report())
    return True


if __name__=="__main__":
    import os
    cache = os.getenv("pydna_cache")
    os.environ["pydna_cache"]="nocache"
    import doctest
    doctest.testmod()
    os.environ["pydna_cache"]=cache



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import pytest
import sys

def test_ipynb_import():

    import mymodule

    assert mymodule.foo2() == "bar2"

    try:
        import IPython
    except ImportError:
        print("*** IPython not installed Jupyter notebook import not tested ***")
        assert True
    else:
        from pydna import ipynb_importer
        import mynotebook
        assert mynotebook.foo() == "bar"

if __name__ == '__main__':
    pytest.cmdline.main([__file__, "-v", "-s"])



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os, re, shutil, pathlib

#for f in [f for f in os.listdir(".") if f.startswith("test_")]:
#    shutil.copy(f, str(pathlib.Path(f).with_suffix(".py_")))


for f in [f for f in os.listdir(".") if f.endswith(".py_")]:
    if not "test_parse.py_" in f:
        continue
    
    print(f)
    
    with open(f,"r") as file_:
        lines = file_.readlines()
    
    newlines=[]
    newline=""
    flag=False
    
    for line in lines:
        if "assert_equal" in line:    
            regex = "(\s+)assert_equal\((.+),(.+)\)"            
            m = re.search(regex, line)            
            newline = "{}assert {} == {}".format(m.groups()[0], m.groups()[1].strip(), m.groups()[2].strip())
            newlines.append(newline)
            flag=True
        else:
            newlines.append(line)
    if flag:        
        nf = str(pathlib.Path(f).with_suffix(".py"))
        print("".join(newlines))
        #import sys; sys.exit(42)
        with open(nf,"w+") as file_:
            file_.writelines(newlines)
    else:
        shutil.copy(f, str(pathlib.Path(f).with_suffix(".py")))


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright 2013-1016 by Björn Johansson. All rights reserved.
# This code is part of the pydna distribution and governed by its
# license.  Please see the LICENSE.txt file that should have been included
# as part of this package.

from .dseqrecord import Dseqrecord
from ._pretty import pretty_str as _ps

class GenbankRecord(Dseqrecord):

    def __init__(self, record, *args, item="accession", start=None, stop=None, strand=1,**kwargs):
        super().__init__(record, *args, **kwargs)
        self.item = item
        self.start = start
        self.stop = stop
        self.strand = strand
        self._repr = item
        if self.start != None and self.stop != None:
            self._repr += " {}-{}".format(self.start, self.stop)
        link = "<a href='https://www.ncbi.nlm.nih.gov/nuccore/{item}?from={start}&to={stop}&strand={strand}' target='_blank'>{text}</a>"
        self.hyperlink = _ps(link.format(item=self.item, start=self.start or "", stop=self.stop or "", strand=self.strand, text=self._repr))

    def __repr__(self):
        '''returns a short string representation of the object'''
        return "Gbank({})({}{})".format(self._repr, {True:"-", False:"o"}[self.linear],len(self))
        
    def _repr_pretty_(self, p, cycle):
        '''returns a short string representation of the object'''
        p.text("Gbank({})({}{})".format(self._repr, {True:"-", False:"o"}[self.linear],len(self)))
            
    def _repr_html_(self):        
        return self.hyperlink



import pydna

from Bio.SeqIO import read as BPread
from Bio.SeqIO import parse as BPparse

q = BPread("read1.gb", "gb")
w = BPread("read2.gb", "gb")
e = BPread("read3.fasta", "fasta")
r = BPread("read4.fasta", "fasta")


with open("pth1.txt", "r", encoding="utf-8") as f:
    a, b = BPparse(f, "gb")

print("|"+a.features[13].qualifiers['label'][0]+"|")
print("|"+a.format("gb")[3314:3324]+"|")

assert a.features[13].qualifiers['label'][0] == '2micron 2µ'
assert a.format("gb")[3314:3324] == '2micron 2µ'

x, y = pydna.parse("pth1.txt")

assert "".join(a.format("gb").splitlines()[1:]) == "".join(x.format("gb").splitlines()[1:])
assert "".join(b.format("gb").strip().splitlines()[4:]) == "".join(y.format("gb").splitlines()[4:])



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import versioneer

# Read author etc. from __init__.py
for line in open('pydna/__init__.py'):
    if line.startswith('__') and not line.startswith('__version') and not line.startswith('__long'):
        exec(line.strip())

from setuptools import setup
try:
    from pypandoc import convert
    read_md = lambda f: convert(f, 'rst')
except ImportError:
    print("warning: pypandoc module not found, could not convert Markdown to RST")
    read_md = lambda f: open(f, 'r').read()

setup(  name            = 'pydna',
        version=versioneer.get_version()[:5],
        cmdclass=versioneer.get_cmdclass(),
        author          = __author__,
        author_email    = __email__,
        zip_safe = True,
        packages=['pydna',
                  'pydna._py_rstr_max',],
        url='http://pypi.python.org/pypi/pydna/',
        license='LICENSE.txt',
        description='''Contains classes and code for representing double
                     stranded DNA and functions for simulating homologous
                     recombination between DNA molecules.''',
        long_description=read_md('README.md'),
        setup_requires=['pytest-runner', "appdirs", "biopython", "prettytable",  "networkx", "ordered-set", "pyparsing"],
        tests_require=['pytest', "appdirs", "biopython", "prettytable",  "networkx", "ordered-set", "pyparsing"],
        install_requires = [ "appdirs", "biopython", "prettytable",  "networkx", "ordered-set", "pyparsing"],
        keywords = "bioinformatics",
        classifiers = ['Development Status :: 4 - Beta',
                       'Environment :: Console',
                       'Intended Audience :: Education',
                       'Intended Audience :: Developers',
                       'Intended Audience :: Science/Research',
                       'License :: OSI Approved :: BSD License',
                       'Operating System :: OS Independent',
                       'Programming Language :: Python :: 3.5',
                       'Topic :: Education',
                       'Topic :: Scientific/Engineering :: Bio-Informatics',])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os            as _os
import sys           as _sys
import subprocess    as _subprocess
import errno         as _errno
import glob          as _glob
from ._pretty        import pretty_str as _pretty_str

'''
# pydna

The pydna package.

:copyright: Copyright 2013 - 2016 by Björn Johansson. All rights reserved.
:license:   This code is part of the pydna distribution and governed by its
            license.  Please see the LICENSE.txt file that should have been included
            as part of this package.

'''

__author__       = "Björn Johansson"
__copyright__    = "Copyright 2013 - 2015 Björn Johansson"
__credits__      = ["Björn Johansson", "Mark Budde"]
__license__      = "BSD"
__maintainer__   = "Björn Johansson"
__email__        = "bjorn_johansson@bio.uminho.pt"
__status__       = "Development" # "Production" #"Prototype"
from ._version import get_versions
__version__      = get_versions()['version'][:5]
__long_version__ = get_versions()['version']
del get_versions


'''
Pydna caches results from the assembly dsdna and amplify
modules. pydna sets an environmental variable "pydna_cache"
which can have three different values:

"cached"        A cache directory if created.
                if possible, cached results are returned.
                if cached results are not available,
                new results are created and cached.
                This is the default.

"nocache"       Results are not written or read from cache.
                No cache directory is created.

"refresh"       new results are made and old are overwritten if they
                exist.

"compare"       The compare functionality is not implemented yet.
                Results are made new and compared with cached results.
                If cached results are not available an exception is raised.

                If new results are not identical with cached, a warning is raised
                and details written to a log located in the data_dir
                The log entry should give the name of the calling script
                if possible and as much details as possible.
                http://victorlin.me/posts/2012/08/26/good-logging-practice-in-python/
'''

import os            as _os
import appdirs       as _appdirs
from configparser import SafeConfigParser as _SafeConfigParser

# create config directory
_os.environ["pydna_config_dir"] = _os.getenv("pydna_config_dir") or _appdirs.user_config_dir("pydna")
try:
    _os.makedirs( _os.environ["pydna_config_dir"] )
except OSError:
    if not _os.path.isdir( _os.environ["pydna_config_dir"] ):
        raise

# set path for the pydna.ini file
_ini_path = _os.path.join( _os.environ["pydna_config_dir"], "pydna.ini" )

# initiate a config parser instance
_parser = _SafeConfigParser()

# if a pydna.ini exists, it is read
if _os.path.exists(_ini_path):
    _parser.read(_ini_path)
else: # otherwise it is created with default settings
    with open(_ini_path, 'w') as f:
        _parser.add_section('main')
        _parser.set('main','email', "someone@example.com")
        _parser.set('main','data_dir', _appdirs.user_data_dir("pydna"))
        _parser.set('main','log_dir',  _appdirs.user_log_dir("pydna"))
        _parser.set('main','cache','cached')
        _parser.set('main','ape','put/path/to/ape/here')
        _parser.set('main','primers','')
        _parser.write(f)

# Six pydna related environmental variables are set from pydna.ini if they are not set already
_os.environ["pydna_email"]    = _os.getenv("pydna_email")    or _parser.get("main", "email")
_os.environ["pydna_data_dir"] = _os.getenv("pydna_data_dir") or _parser.get("main", "data_dir")
_os.environ["pydna_log_dir"]  = _os.getenv("pydna_log_dir")  or _parser.get("main", "log_dir")
_os.environ["pydna_cache"]    = _os.getenv("pydna_cache")    or _parser.get("main", "cache")
_os.environ["pydna_ape"]      = _os.getenv("pydna_ape")      or _parser.get("main", "ape")
_os.environ["pydna_primers"]  = _os.getenv("pydna_primers")  or _parser.get("main", "primers")

# Check sanity of pydna_cache variable
if _os.environ["pydna_cache"] not in ("cached", "nocache", "refresh", "compare"):
    raise Exception("cache (os.environ['pydna_cache']) is not cached, nocache, refresh or compare")

# create log directory if not present
try:
    _os.makedirs( _os.environ["pydna_log_dir"] )
    _logmsg = "Created log directory {}".format(_os.environ["pydna_log_dir"])
except OSError:
    if _os.path.isdir( _os.environ["pydna_log_dir"] ):
        _logmsg = "Log directory {} found.".format(_os.environ["pydna_log_dir"])
    else:
        raise

# create logger
import logging as _logging
import logging.handlers as _handlers
_logger = _logging.getLogger("pydna")
#_logger.setLevel(_logging.DEBUG)
_hdlr = _handlers.RotatingFileHandler(_os.path.join( _os.environ["pydna_log_dir"] , 'pydna.log'), mode='a', maxBytes=10*1024*1024, backupCount=10, encoding='utf-8')
_formatter = _logging.Formatter('%(asctime)s %(levelname)s %(message)s')
_hdlr.setFormatter(_formatter)
_logger.addHandler(_hdlr)
_logger.info(_logmsg)

_logger.info('Assigning environmental variable pydna_data_dir = {}'.format( _os.environ["pydna_data_dir"] ))

# create cache directory if not present
try:
    _os.makedirs( _os.environ["pydna_data_dir"] )
    _logger.info("Created data directory {}".format(_os.environ["pydna_data_dir"]))
except OSError:
    if _os.path.isdir( _os.environ["pydna_data_dir"] ):
        _logger.info("data directory {} found".format(_os.environ["pydna_data_dir"]))
    else:
        raise

#try:
#    _os.makedirs( _os.environ["pydna_cache"] )
#    print(234)
#    _logger.info("Created cache directory {}".format(_os.environ["pydna_cache"]))
#except OSError:
#    if _os.path.isdir( _os.environ["pydna_cache"] ):
#        _logger.info("data directory {} found".format(_os.environ["pydna_cache"]))
#    else:
#        raise
        
from .amplify    import Anneal
from .amplify    import pcr
from .amplify    import nopcr
from .assembly   import Assembly
from .genbank    import Genbank
from .genbank    import genbank
from .download   import download_text
from .dseq       import Dseq
from .dseqrecord import Dseqrecord
from .parsers    import parse
from .readers    import read
from .parsers    import parse_primers
from .readers    import read_primer

#from .editor                             import Editor
from .findsubstrings_suffix_arrays_python import common_sub_strings
from .primer_design                       import cloning_primers
from .primer_design                       import assembly_primers
from .primer_design                       import integration_primers
from .utils                               import eq
#from .utils                               import shift_origin
#from .utils                               import pairwise
#from .utils                               import cseguid
#from .primer                              import Primer
#from .genbankfixer                        import gbtext_clean

# find out if optional dependecies for gel module are in place
_missing_modules_for_gel = []

try:
    import scipy
    del scipy
except ImportError:
    _missing_modules_for_gel.append("scipy")
try:
    import numpy
    del numpy
except ImportError:
    _missing_modules_for_gel.append("numpy")
try:
    import matplotlib
    del matplotlib
except ImportError:
    _missing_modules_for_gel.append("matplotlib")
try:
    import mpldatacursor
    del mpldatacursor
except ImportError:
    _missing_modules_for_gel.append("mpldatacursor")
try:
    import pint
    del pint
except ImportError:
    _missing_modules_for_gel.append("pint")

if _missing_modules_for_gel:
    _logger.warning("gel simulation will NOT be available. Missing modules: {}"
        .format(", ".join(_missing_modules_for_gel)))
else:
    from .gel import Gel

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions

class PydnaWarning(Warning):
    """Pydna warning.

    Pydna uses this warning (or subclasses of it), to make it easy to
    silence all warning messages:

    >>> import warnings
    >>> from pydna import PydnaWarning
    >>> warnings.simplefilter('ignore', PydnaWarning)

    Consult the warnings module documentation for more details.
    """
    pass


class PydnaDeprecationWarning(PydnaWarning):
    """pydna deprecation warning.

    Pydna uses this warning instead of the built in DeprecationWarning
    since those are ignored by default since Python 2.7.

    To silence all our deprecation warning messages, use:

    >>> import warnings
    >>> from pydna import PydnaDeprecationWarning
    >>> warnings.simplefilter('ignore', PydnaDeprecationWarning)

    Code marked as deprecated will be removed in a future version
    of Pydna. This can be discussed in the Pydna google group:
    https://groups.google.com/forum/#!forum/pydna    
    
    """
    pass

try:
    del amplify
except NameError:
    pass
try:
    del assembly
except NameError:
    pass
try:
    del download
except NameError:
    pass
try:
    del dseq
except NameError:
    pass
try:
    del dseqrecord
except NameError:
    pass
try:
    del readers
except NameError:
    pass
try:
    del parsers
except NameError:
    pass
try:
    del _findsubstrings_suffix_arrays_python
except NameError:
    pass
from .genbank    import genbank
try:
    del primer_design
except NameError:
    pass
try:
    del utils
except NameError:
    pass
try:
    del genbankfixer
except NameError:
    pass


def open_cache_folder():
    _open_folder( _os.environ["pydna_data_dir"] )

def open_config_folder():
    _open_folder( _os.environ["pydna_config_dir"] )

def open_log_folder():
    _open_folder( _os.environ["pydna_log_dir"] )

def _open_folder(pth):
    if _sys.platform=='win32':
        _subprocess.Popen(['start', pth], shell= True)
    elif _sys.platform=='darwin':
        _subprocess.Popen(['open', pth])
    else:
        try:
            _subprocess.Popen(['xdg-open', pth])
        except OSError:
            return "no cache to open."
            
def delete_cache(categories=[ "amplify*", "assembly*", "genbank*", "web*", "synced*" ]):
    msg = "cache file deletion\n"
    for cat in categories:
        files = _glob.glob(_os.path.join( _os.environ["pydna_data_dir"], cat))
        for file_ in files:
            msg += file_
            try:
                _os.remove( file_ )
                msg += " deleted.\n"
            except OSError as e:
                if e._errno == _errno.ENOENT:
                    msg += " no file to delete.\n"
    return _pretty_str(msg)
   
def nocache():
    _os.environ["pydna_cache"]="nocache"
def cached():
    _os.environ["pydna_cache"]="cached"
def refresh():
    _os.environ["pydna_cache"] ="refresh"
    
def getcache():
    return _pretty_str( _os.getenv("pydna_cache") )


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright 2013 by Björn Johansson.  All rights reserved.
# This code is part of the Python-dna distribution and governed by its
# license.  Please see the LICENSE.txt file that should have been included
# as part of this package.

from Bio.SeqRecord import SeqRecord

class SeqRecord(SeqRecord):
    pass 



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# doctest: +NORMALIZE_WHITESPACE
# doctest: +SKIP
'''This module provides functions for PCR. Primers with 5' tails as well as inverse PCR on
circular templates are handled correctly.

'''
import math
import itertools

import re
import textwrap
import copy

import logging
module_logger = logging.getLogger("pydna."+__name__)

from Bio.SeqRecord                  import SeqRecord
from Bio.SeqUtils                   import GC
from Bio.SeqUtils.MeltingTemp       import Tm_NN
from .utils                        import rc
from .dseqrecord                   import Dseqrecord
from ._pretty                       import pretty_str, pretty_unicode
from .tm                           import tmbresluc

def _annealing_positions(primer, template, limit=15):
    '''Finds the annealing position(s) for a primer on a template where the
    primer anneals perfectly with at least limit nucleotides in the 3' part.

    start is a position (integer)
    footprint1 and tail1 are strings.

    ::

        <- - - - - - - - - - template - - - - - - - - - - - - - >

        <------------- start ---->
     5'-...gctactacacacgtactgactgcctccaagatagagtcagtaaccacactcgat...3'
           ||||||||||||||||||||||||||||||||||||||||||||||||
                                  3'-gttctatctcagtcattggtgtATAGTG-5'

                                                        <tail>
                                     <---footprint----->
                                     <--------- primer ------>

    Parameters
    ----------
    primer : string
        The primer sequence 5'-3'

    template : string
        The template sequence 5'-3'

    limit : int = 15, optional
        footprint needs to be at least of length limit.

    Returns
    -------
    describe : list of tuples (int, string, string)
        [ (start1, footprint1, tail1), (start2, footprint2, tail2),..., ]
    '''

    if len(primer)<limit:
        return []
    prc = rc(primer)
    head = prc[:limit].upper()

    table = {"R":"(A|G)",
             "Y":"(C|T)",
             "S":"(G|C)",
             "W":"(A|T)",
             "K":"(G|T)",
             "M":"(A|C)",
             "B":"(C|G|T)",
             "D":"(A|G|T)",
             "H":"(A|C|T)",
             "V":"(A|C|G)",
             "N":"(A|G|C|T)"}

    for key in table:
        head=head.replace(key, table[key])

    positions = [m.start() for m in re.finditer('(?={})'.format(head), template, re.I)]

    if positions:
        tail = prc[limit:]
        length = len(tail)
        results = []
        for match_start in positions:
            tm = template[match_start+limit:match_start+limit+length]
            footprint = rc(template[match_start:match_start+limit]+"".join([b for a,b in itertools.takewhile(lambda x: x[0].lower()==x[1].lower(), list(zip(tail, tm)))]))
            results.append((match_start, footprint, primer[: len(primer) - len(footprint) ]))
        return results
    return []

class Amplicon(Dseqrecord):
    '''The Amplicon class holds information about a PCR reaction involving two
    primers and one template. This class is used by the Anneal class and is not
    meant to be instantiated directly.

    Parameters
    ----------
    forward_primer : SeqRecord(Biopython)
        SeqRecord object holding the forward (sense) primer

    reverse_primer : SeqRecord(Biopython)
        SeqRecord object holding the reverse (antisense) primer

    template : Dseqrecord
        Dseqrecord object holding the template (circular or linear)

    saltc : float, optional
        saltc = monovalent cations (mM) (Na,K..)
        default value is 50mM
        This is used for Tm calculations.

    forward_primer_concentration : float, optional
        primer concentration (nM)
        default set to 1000nM = 1µM
        This is used for Tm calculations.

    rc : float, optional
        primer concentration (nM)
        default set to 1000nM = 1µM
        This is used for Tm calculations.
    '''

    def __init__(    self,
                     record,
                     *args,
                     template=None,
                     forward_primer=None,
                     reverse_primer=None,
                     saltc=None,
                     fprimerc=1000.0,
                     rprimerc=1000.0,
                     **kwargs):

        super().__init__(record, *args, **kwargs)
        self.template = template
        self.forward_primer = forward_primer
        self.reverse_primer = reverse_primer
        self.fwd_primer = self.forward_primer
        self.rev_primer = self.reverse_primer
        self.fprimerc = fprimerc
        self.rprimerc = rprimerc
        self.saltc = saltc

    def __getitem__(self, sl):
        answer = copy.copy(self)
        answer.seq = answer.seq.__getitem__(sl)
        answer.seq.alphabet = self.seq.alphabet
        sr = SeqRecord("n"*len(self))
        sr.features = self.features
        answer.features = SeqRecord.__getitem__(sr, sl).features
        return answer

    def __repr__(self):
        '''returns a short string representation of the object'''
        return "Amplicon({})".format(self.__len__())

    def _repr_pretty_(self, p, cycle):
            p.text("Amplicon({})".format(self.__len__()))
            
    def _repr_html_(self):
        return "Amplicon({})".format(self.__len__())

    def flankup(self, flankuplength=50):
        '''Returns a Dseqrecord object containing flankuplength bases upstream of the forward primer footprint,
       Truncated if the template is not long enough.

       ::

        <--- flankup --->

                  5TAATAAactactgactatct3
                         ||||||||||||||
        acgcattcagctactgtactactgactatctatcg

       '''
        return self.template.seq[self.forward_primer.position-flankuplength-len(self.forward_primer.footprint):self.forward_primer.position-len(self.forward_primer.footprint)]

    def flankdn(self, flankdnlength=50):
        '''Returns a Dseqrecord object containing flankdnlength bases downstream of the reverse primer footprint.
       Truncated if the template is not long enough.

       ::

                                       <---- flankdn ------>

                        3actactgactatctTAATAA5
                         ||||||||||||||
        acgcattcagctactgtactactgactatctatcgtacatgtactatcgtat


       '''
        return self.template.seq[self.reverse_primer.position+len(self.reverse_primer.footprint):self.reverse_primer.position+flankdnlength+len(self.reverse_primer.footprint)]



    def figure(self):
        '''
        This method returns a simple figure of the two primers binding to a part
        of the template.

        ::

         5gctactacacacgtactgactg3
          |||||||||||||||||||||| tm 52.6 (dbd) 58.3
         5gctactacacacgtactgactg...caagatagagtcagtaaccaca3
         3cgatgatgtgtgcatgactgac...gttctatctcagtcattggtgt5
                                   |||||||||||||||||||||| tm 49.1 (dbd) 57.7
                                  3gttctatctcagtcattggtgt5



        Returns
        -------
        figure:string
             A string containing a text representation of the primers
             annealing on the template (see example above).


        Notes
        -----
        tm in the figure above is the melting temperature (tm) for each primer calculated according to
        SantaLucia 1998 [#]_.

        dbd is the tm calculation for enzymes with dsDNA binding domains like Pfu-Sso7d [#]_. See [#]_
        for more information.

        References
        ----------

        .. [#] J. SantaLucia, “A Unified View of Polymer, Dumbbell, and Oligonucleotide DNA Nearest-neighbor Thermodynamics,” Proceedings of the National Academy of Sciences 95, no. 4 (1998): 1460.
        .. [#] M. Nørholm, “A Mutant Pfu DNA Polymerprimerase Designed for Advanced Uracil-excision DNA Engineering,” BMC Biotechnology 10, no. 1 (2010): 21, doi:10.1186/1472-6750-10-21.
        .. [#] http://www.thermoscientificbio.com/webtools/tmc/

        '''

        tmf = Tm_NN(str(self.forward_primer.footprint),
                    dnac1=self.fprimerc,
                    Na=self.saltc)
        tmr = Tm_NN(str(self.reverse_primer.footprint),
                    dnac1=self.fprimerc,
                    Na=self.saltc)

        tmf_dbd = tmbresluc(str(self.forward_primer.footprint), primerc=self.fprimerc)
        tmr_dbd = tmbresluc(str(self.reverse_primer.footprint), primerc=self.rprimerc)

        f =   '''
            {sp1}5{faz}...{raz}3
             {sp3}{rap} tm {tmr} (dbd) {tmr_dbd}
            {sp3}3{rp}5
            5{fp}3
             {fap:>{fplength}} tm {tmf} (dbd) {tmf_dbd}
            {sp2}3{fzc}...{rzc}5
            '''.format( fp       = self.forward_primer.seq,
                        fap      = "|"*len(self.forward_primer.footprint),
                        fplength = len(self.forward_primer.seq),
                        tmf      = round(tmf,1),
                        tmr      = round(tmr,1),
                        tmf_dbd  = round(tmf_dbd,1),
                        tmr_dbd  = round(tmr_dbd,1),
                        rp       = self.reverse_primer.seq[::-1],
                        rap      = "|"*len(self.reverse_primer.footprint),
                        rplength = len(self.reverse_primer.seq),
                        faz      = self.forward_primer.footprint,
                        raz      = self.reverse_primer.footprint.reverse_complement(),
                        fzc      = self.forward_primer.footprint.complement(),
                        rzc      = self.reverse_primer.footprint[::-1],
                        sp1      = " "*(len(self.forward_primer.seq)-len(self.forward_primer.footprint)),
                        sp2      = " "*(len(self.forward_primer.seq)-len(self.forward_primer.footprint)),
                        sp3      = " "*(3+len(self.forward_primer.seq))
                       )
        return pretty_str(textwrap.dedent(f).strip("\n"))


    def program(self):

        '''Returns a string containing a text representation of a proposed
       PCR program using Taq or similar polymerase.

       ::

        Taq (rate 30 nt/s)
        Three-step|         30 cycles     |      |Tm formula: Biopython Tm_NN
        94.0°C    |94.0°C                 |      |SaltC 50mM
        __________|_____          72.0°C  |72.0°C|
        04min00s  |30s  \         ________|______|
                  |      \ 46.0°C/ 0min 1s|10min |
                  |       \_____/         |      |
                  |         30s           |      |4-8°C

       '''


        # Primer melting temperatures are calculated with the Tm_NN formula from
        # biopython
        # simple salt concentration correction is used and the template concentration
        # is ignored. dnac1 = primer concentration
        
        tmf = Tm_NN(str(self.forward_primer.footprint),
                    dnac1=self.fprimerc,
                    Na=self.saltc)
        tmr = Tm_NN(str(self.reverse_primer.footprint),
                    dnac1=self.fprimerc,
                    Na=self.saltc)

        # Ta calculation according to
        # Rychlik, Spencer, and Rhoads, 1990, Optimization of the anneal
        # ing temperature for DNA amplification in vitro
        # http://www.ncbi.nlm.nih.gov/pubmed/2243783
        # The formula described uses the length and GC content of the product and
        # salt concentration (monovalent cations).

        #GC_prod=GC(str(self.seq))

        tmp = 81.5 + 0.41*GC(str(self.seq)) + 16.6*math.log10(self.saltc/1000.0) - 675/len(self)
        tml = min(tmf,tmr)
        ta = 0.3*tml+0.7*tmp-14.9

        # Taq polymerase extension rate is set to 30 nt/s
        # see https://www.thermofisher.com/pt/en/home/life-science/pcr/pcr-enzymes-master-mixes/taq-dna-polymerase-enzymes/taq-dna-polymerase.html
        taq_extension_rate = 30  # seconds/kB PCR product length
        extension_time_taq = int(round(taq_extension_rate * len(self) / 1000)) # seconds
        f  = textwrap.dedent('''
                                 Taq (rate {rate} nt/s) 35 cycles             |{size}bp
                                 95.0°C    |95.0°C                 |      |Tm formula: Biopython Tm_NN
                                 |_________|_____          72.0°C  |72.0°C|SaltC {saltc:2}mM
                                 | 03min00s|30s  \         ________|______|Primer1C {forward_primer_concentration:3}µM
                                 |         |      \ {ta}°C/{0:2}min{1:2}s| 5min |Primer2C {reverse_primer_concentration:3}µM
                                 |         |       \_____/         |      |GC {GC_prod}%
                                 |         |         30s           |      |4-12°C'''.format(rate=taq_extension_rate,
                                                                                            forward_primer_concentration=self.fprimerc/1000,
                                                                                            reverse_primer_concentration=self.rprimerc/1000,
                                                                                            ta=round(ta,1),
                                                                                            saltc=self.saltc,
                                                                                            *divmod(extension_time_taq,60),
                                                                                            size= len(self.seq),
                                                                                            GC_prod= int(self.gc()) ))
        return pretty_unicode(f)

    def taq_program(self):
        return self._program()

    def dbd_program(self):
        '''Returns a string containing a text representation of a proposed
       PCR program using a polymerase with a DNA binding domain such as Pfu-Sso7d.

       ::

        Pfu-Sso7d (rate 15s/kb)             |{size}bp
        Three-step|          30 cycles   |      |Tm formula: Pydna tmbresluc
        98.0°C    |98.0°C                |      |SaltC 50mM
        __________|_____          72.0°C |72.0°C|Primer1C   1µM
        00min30s  |10s  \ 61.0°C ________|______|Primer2C   1µM
                  |      \______/ 0min 0s|10min |
                  |        10s           |      |4-12°C

       '''
        PfuSso7d_extension_rate = 15                #seconds/kB PCR product
        extension_time_PfuSso7d = PfuSso7d_extension_rate * len(self) / 1000  # seconds

        # The program returned is eaither a two step or three step progrem
        # This depends on the tm and length of the primers in the
        # original instructions from finnzyme. These do not seem to be

        tmf_dbd = tmbresluc(str(self.forward_primer.footprint), primerc=self.fprimerc)
        tmr_dbd = tmbresluc(str(self.reverse_primer.footprint), primerc=self.rprimerc)

        # Ta calculation for enzymes with dsDNA binding domains like Pfu-Sso7d
        # https://www.finnzymes.fi/tm_determination.html

        length_of_f = len(self.forward_primer.footprint)
        length_of_r = len(self.reverse_primer.footprint)

        if (length_of_f>20 and length_of_r>20 and tmf_dbd>=69.0 and tmr_dbd>=69.0) or (tmf_dbd>=72.0 and tmr_dbd>=72.0):
            f=textwrap.dedent(  '''
                                    Pfu-Sso7d (rate {rate}s/kb)
                                    Two-step|    30 cycles |      |{size}bp
                                    98.0°C  |98.0C         |      |Tm formula: Pydna tmbresluc
                                    _____ __|_____         |      |SaltC {saltc:2}mM
                                    00min30s|10s  \        |      |Primer1C {forward_primer_concentration:3}µM
                                            |      \ 72.0°C|72.0°C|Primer2C {reverse_primer_concentration:3}µM
                                            |       \______|______|GC {GC_prod}%
                                            |      {0:2}min{1:2}s|10min |4-12°C
                                 '''.format(rate = PfuSso7d_extension_rate,
                                            forward_primer_concentration = self.fprimerc/1000,
                                            reverse_primer_concentration = self.rprimerc/1000,
                                            saltc = self.saltc,
                                            *map(int,divmod(extension_time_PfuSso7d,60)),
                                            GC_prod= int(self.gc()),
                                            size = len(self.seq) ))
        else:

            if (length_of_f>20 and length_of_r>20):
                ta = min(tmf_dbd,tmr_dbd)+3
            else:
                ta = min(tmf_dbd,tmr_dbd)

            f=textwrap.dedent(  '''
                                    Pfu-Sso7d (rate {rate}s/kb)                 |{size}bp
                                    Three-step|          30 cycles   |      |Tm formula: Pydna tmbresluc
                                    98.0°C    |98.0°C                |      |SaltC {saltc:2}mM
                                    __________|_____          72.0°C |72.0°C|Primer1C {forward_primer_concentration:3}µM
                                    00min30s  |10s  \ {ta:.1f}°C ________|______|Primer2C {reverse_primer_concentration:3}µM
                                              |      \______/{0:2}min{1:2}s|10min |GC {GC_prod}%
                                              |        10s           |      |4-12°C
                                 '''.format(rate = PfuSso7d_extension_rate,
                                            size= len(self.seq),
                                            ta   = round(ta),
                                            forward_primer_concentration   = self.fprimerc/1000,
                                            reverse_primer_concentration   = self.rprimerc/1000,
                                            saltc= self.saltc,
                                            GC_prod= int(self.gc()),
                                            *map(int, divmod(extension_time_PfuSso7d,60)) ))
        return pretty_str(f)

    def pfu_sso7d_program(self):
        return self.dbd_program()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
'''
test parse
'''

import pytest
import sys

from pydna import parse
from pydna import pcr
from Bio.SeqUtils.CheckSum import seguid

def test_circ_pcr():
    ''' test circ pcr'''
    s = parse('''
    >MCT4_flaghis_rv
    gccgcaagcttgtcgtcatcgtctttgtagtcCATggccaggagggttggttccgcc
    >MCT4_flaghis_fw
    ccaccaccaccaccaccaccaccaccaccacggaggggccgtggtggacgagggcc''', ds=False)

    t = parse('''
    >hej circular
    TCAGGTGAGGCGGAACCAACCCTCCTGGCCATGggaggggccgtggtggacgagggccccacaggcg
    ''')

    p = pcr(s,t)

    assert p.seguid() == 'wYEpbWxO8tfgA_AL67wqGzYmc5k'



def test_pcr():
    ''' test pcr'''

    raw=[]

    raw.append(

    ('7JOV1MJBZJp2Smja/7KFGhS2SWY',

    parse('''
    >524_pFA6aF (29-mer)
    cacatacgatttaggtgacactatagaac

    >523_AgTEF1tpR (21-mer)
    ggttgtttatgttcggatgtg
    '''),

    parse("pAG25.gb"),)
    )

    raw.append(

    ('7pPxy/bQvs4+7CaOgiywQVzUFDc',

    parse('''
    >lowgc_f
    TTTCACTAGTTACTTGTAGTCGacgtgccatctgtgcagacaaacgcatcaggatat

    >lowgc_r
    AAGTTGGAAATCTAGCTTTTCTTgacgtcagcggccgcattgcaca
    '''),
    parse("pCAPs.gb"),)
    )

    raw.append(

    ('7JOV1MJBZJp2Smja/7KFGhS2SWY',

    parse('''
    >524_pFA6aF (29-mer)
    cacatacgatttaggtgacactatagaac

    >523_AgTEF1tpR (21-mer)
    ggttgtttatgttcggatgtg
    '''),

    parse("pAG25.gb"),

    ))

    raw.append(

    ('yshvYTXr9iXCnh3YytWQRDBNQzI',

    parse('''
    >ForwardPrimer1
    gctactacacacgtactgactg

    >ReversePrimer
    tgtggttactgactctatcttg

    LOCUS       sequence_50_bp            46 bp    DNA     circular UNK 08-FEB-2013
    DEFINITION  sequence_50_bp circular
    ACCESSION   sequence_50_bp
    VERSION     sequence_50_bp
    KEYWORDS    .
    SOURCE      .
      ORGANISM  .
                .
    FEATURES             Location/Qualifiers
    ORIGIN
            1 ccaagataga gtcagtaacc acagctacta cacacgtact gactgt
    //

    '''),))

    raw.append(

    ('yshvYTXr9iXCnh3YytWQRDBNQzI',

    parse('''
    >ForwardPrimer2
    gctactacacacgtactgactg

    >ReversePrimer
    tgtggttactgactctatcttg

    LOCUS       template                  46 bp    DNA     circular UNK 15-OCT-2012
    DEFINITION  template circular
    ACCESSION   template
    VERSION     template
    KEYWORDS    .
    SOURCE      .
      ORGANISM  .
                .
    FEATURES             Location/Qualifiers
    ORIGIN
            1 ccaagataga gtcagtaacc acagctacta cacacgtact gactgt
    //
    '''),))
    raw.append(

    ('yshvYTXr9iXCnh3YytWQRDBNQzI',

    parse('''
    >ForwardPrimer3
    gctactacacacgtactgactg

    >ReversePrimer
    tgtggttactgactctatcttg

    LOCUS       template                  46 bp    DNA     circular UNK 08-FEB-2013
    DEFINITION  template circular
    ACCESSION   template
    VERSION     template
    KEYWORDS    .
    SOURCE      .
      ORGANISM  .
                .
    FEATURES             Location/Qualifiers
    ORIGIN
            1 tccaagatag agtcagtaac cacagctact acacacgtac tgactg
    //
    '''),))

    raw.append(
    ('yshvYTXr9iXCnh3YytWQRDBNQzI',

    parse('''
    >ForwardPrimer4
    gctactacacacgtactgactg

    >ReversePrimer
    tgtggttactgactctatcttg

    LOCUS       template                  46 bp    DNA     circular UNK 15-OCT-2012
    DEFINITION  template circular
    ACCESSION   template
    VERSION     template
    KEYWORDS    .
    SOURCE      .
      ORGANISM  .
                .
    FEATURES             Location/Qualifiers
    ORIGIN
            1 gtccaagata gagtcagtaa ccacagctac tacacacgta ctgact
    //
    '''),))

    raw.append(
    ('60meNXeGKO7ahZwcIl5yXHFC3Yg',
    parse('''
    >fw1
    cacatacgatttaggtgacactatagaac
    >rv
    ggttgtttatgttcggatgtg

    LOCUS       tm                        50 bp    DNA     circular UNK 15-OCT-2012
    DEFINITION  tm circular
    ACCESSION   tm
    VERSION     tm
    KEYWORDS    .
    SOURCE      .
      ORGANISM  .
                .
    FEATURES             Location/Qualifiers
    ORIGIN
            1 cacatccgaa cataaacaac ccacatacga tttaggtgac actatagaac
    //
    '''),)
    )

    raw.append(

    ('60meNXeGKO7ahZwcIl5yXHFC3Yg',
    parse('''
    >fw2
    cacatacgatttaggtgacactatagaac
    >rv
    ggttgtttatgttcggatgtg

    LOCUS       tm                        50 bp    DNA     circular UNK 15-OCT-2012
    DEFINITION  tm circular
    ACCESSION   tm
    VERSION     tm
    KEYWORDS    .
    SOURCE      .
      ORGANISM  .
                .
    FEATURES             Location/Qualifiers
    ORIGIN
            1 acatccgaac ataaacaacc cacatacgat ttaggtgaca ctatagaacc
    //

    '''),)
    )

    raw.append(

    ('60meNXeGKO7ahZwcIl5yXHFC3Yg',
    parse('''
    >fw3
    cacatacgatttaggtgacactatagaac
    >rv
    ggttgtttatgttcggatgtg
    LOCUS       tm                        50 bp    DNA     circular UNK 15-OCT-2012
    DEFINITION  tm circular
    ACCESSION   tm
    VERSION     tm
    KEYWORDS    .
    SOURCE      .
      ORGANISM  .
                .
    FEATURES             Location/Qualifiers
    ORIGIN
            1 ccacatccga acataaacaa cccacatacg atttaggtga cactatagaa
    //

    '''),)
    )


    raw.append(
    ('y6ohCJ4O+8Is012DItz4F4saxNo',
    parse('''
    >f_Eric_Ma
    ARATGAGTCTTCTRACCGAGGTCG
    >r_Eric_Ma
    TGAAAAGACATCYTCAAGYYTCTG
    >templ
    AGCAAAAGCAGGTAGATATTGAAAAATGAGTCTTCTAACCGAGGTCGAAACGTACGTTCTCTCTATCGTCCCGTCAGGCCCCCTCAAAGCCGAGATCGCGCAGAGACTTGAAGATGTCTCTGCAGGGAAGAACACTGATCTCGAGGCTCTCATGGAATGGCTAAAGACAAGACCAATCCTGTCACCTCTGACTAAGGGGATTTTAGGGTTTGTGTTCACGCTCACCGTGCCCAGTGAGCGAGGACTGCAGCGTAGACGCTTTGTCCAGAATGCCTTAAATGGGAATGGAGACCCAAACAACATGGACAGGGCAGTCAAACTATACAGGAAGCTGAAAAGAGAGATAACATTCCATGGGGCTAAAGAGGTTGCACTCAGCTATTCAACCGGTGCACTTGCCAGTTGCATGGGTCTCATATACAACAGGATGGGAACGGTAACCACAGAAGTAGCTTTTGGCCTGGTGTGTGCCACTTGTGAGCAGATTGCTGACTCACAGCATCGATCTCACAGACAGATGGTGACTACCACCAACCCACTAATCAGGCATGAAAACAGAATGGTGCTGGCCAGCACTACAGCTAAGGCTATGGAGCAGATGGCTGGATCGAGTGAACAGGCAGCGGAAGCCATGGAGGTTGCTAGTCAGGCTAGGCAGATGGTGCAGGCAATGAGGACAATTGGGACTCACCCTAGCTCCAGTGCCGGTCTGAAAGATGATCTTCTTGAAAATTTGCAGGCCTACCAGAAGCGGATGGGAGTGCAAATGCAGCGATTCAAGTGATCCTCTCGTTATTGCCGCAAGTATCATTGGGATCTTGCACTTGATATTGTGGATTCTTGATCGTCCTTTCTTCAAATGTATTTATCGTCGCCTTAAATACGGTTTGAAAAGAGGGCCTTCTACGGAAGGAGTGCCTGAGTCTATGAGGGAAGAGTATCGGCAGGAACAGCAGAGTGCTGTGGATGTTGACGATGGTCATTTTGTCAACATAGAGCTGGAGTAAAAAACTACCTTGTTTCTACT
    '''),)
    )

    for key, tst in enumerate(raw):
        print(key, end=" ")
        print(tst[1][0].name, end=" ")
        print("pcr test", key)
        assert tst[0] == seguid(pcr(tst[1:]).seq)

if __name__ == '__main__':
    pytest.cmdline.main([__file__, "-v", "-s"])
